# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# AUTHORS: Josep Fanals Batllori and Santiago Peñate Vera
# CONTACT:  u1946589@campus.udg.edu and santiago.penate.vera@gmail.com
# thanks to Llorenç Fanals Batllori for his help at coding
import pandas as pd
import numpy as np
import numba as nb
from numba.np.linalg import solve_impl
import time
from warnings import warn
from scipy.sparse import csc_matrix, coo_matrix
from scipy.sparse import hstack as hs, vstack as vs
from scipy.sparse.linalg import spsolve, factorized
from GridCal.Engine.Simulations.PowerFlow.power_flow_results import NumericPowerFlowResults
from GridCal.Engine.Simulations.PowerFlow.NumericalMethods.common_functions import *
from GridCal.Engine.basic_structures import Logger


def epsilon(Sn, n, E):
    """
    Fast recursive Wynn's epsilon algorithm from:
        NONLINEAR SEQUENCE TRANSFORMATIONS FOR THE ACCELERATION OF CONVERGENCE
        AND THE SUMMATION OF DIVERGENT SERIES

        by Ernst Joachim Weniger
    Args:
        Sn: sum of coefficients
        n: order
        E: Coefficients structure copy that is modified in this algorithm

    Returns:

    """
    Zero = complex(0)
    One = complex(1)
    Tiny = np.finfo(complex).tiny  # np.finfo(complex).min
    Huge = np.finfo(complex).max

    E[n] = Sn

    if n == 0:
        estim = Sn
    else:
        AUX2 = Zero

        for j in range(n, 0, -1):  # range from n to 1 (both included)
            AUX1 = AUX2
            AUX2 = E[j - 1]
            DIFF = E[j] - AUX2

            if abs(DIFF) <= Tiny:
                E[j - 1] = Huge
            else:
                if DIFF == 0:
                    DIFF = Tiny
                E[j - 1] = AUX1 + One / DIFF

        if np.mod(n, 2) == 0:
            estim = E[0]
        else:
            estim = E[1]

    return estim, E


# @nb.njit("(c16[:])(i8, c16[:, :], f8)")
def pade4all(order, coeff_mat, s=1.0):
    """
    Computes the "order" Padè approximant of the coefficients at the approximation point s

    Arguments:
        coeff_mat: coefficient matrix (order, buses)
        order:  order of the series
        s: point of approximation (at 1 you get the voltage)

    Returns:
        Padè approximation at s for all the series
    """
    nbus = coeff_mat.shape[1]

    # complex_type = nb.complex128
    complex_type = np.complex128

    voltages = np.zeros(nbus, dtype=complex_type)

    nn = int(order / 2)
    L = nn
    M = nn

    for d in range(nbus):

        # formation of the linear system right hand side
        rhs = coeff_mat[L + 1:L + M + 1, d]

        # formation of the coefficients matrix
        C = np.zeros((L, M), dtype=complex_type)
        for i in range(L):
            k = i + 1
            C[i, :] = coeff_mat[L - M + k:L + k, d]

        # Obtaining of the b coefficients for orders greater than 0
        b = np.zeros(rhs.shape[0] + 1, dtype=complex_type)
        x = np.linalg.solve(C, -rhs)  # bn to b1
        b[0] = 1
        b[1:] = x[::-1]

        # Obtaining of the coefficients 'a'
        a = np.zeros(L + 1, dtype=complex_type)
        a[0] = coeff_mat[0, d]
        for i in range(L):
            val = complex_type(0)
            k = i + 1
            for j in range(k + 1):
                val += coeff_mat[k - j, d] * b[j]
            a[i + 1] = val

        # evaluation of the function for the value 's'
        p = complex_type(0)
        q = complex_type(0)
        for i in range(L + 1):
            p += a[i] * s ** i
            q += b[i] * s ** i

        voltages[d] = p / q

    return voltages


# @nb.njit("(c16[:])(c16[:, :], c16[:, :], i8, c16[:])")
@nb.njit(cache=True)
def sigma_function(coeff_matU, coeff_matX, order, V_slack):
    """

    :param coeff_matU: array with voltage coefficients
    :param coeff_matX: array with inverse conjugated voltage coefficients
    :param order: should be prof - 1
    :param V_slack: slack bus voltage vector. Must contain only 1 slack bus
    :return: sigma complex value
    """
    if len(V_slack) > 1:
        print('Sigma values may not be correct')
    V0 = V_slack[0]
    coeff_matU = coeff_matU / V0
    coeff_matX = coeff_matX / V0
    nbus = coeff_matU.shape[1]
    complex_type = nb.complex128
    # complex_type = complex
    sigmas = np.zeros(nbus, dtype=complex_type)

    if order % 2 == 0:
        M = int(order / 2) - 1
    else:
        M = int(order / 2)

    for d in range(nbus):
        a = coeff_matU[1:2 * M + 2, d]
        b = coeff_matX[0:2 * M + 1, d]
        C = np.zeros((2 * M + 1, 2 * M + 1), dtype=complex_type)

        for i in range(2 * M + 1):
            if i < M:
                C[1 + i:, i] = a[:2 * M - i]
            else:
                C[i - M:, i] = - b[:3 * M - i + 1]

        lhs = np.linalg.solve(C, -a)

        sigmas[d] = np.sum(lhs[M:]) / (np.sum(lhs[:M]) + 1)

    return sigmas


# @nb.njit("(c16[:])(c16[:, :], c16[:, :], i8, i8[:])")
@nb.njit(cache=True)
def conv1_old(A, B, c, indices):
    """
    Performs the convolution of A* and B
    :param A: Coefficients matrix 1 (orders, buses)
    :param B: Coefficients matrix 2 (orders, buses)
    :param c: order of the coefficients
    :param indices: bus indices array
    :return: Array with the convolution for the buses given by "indices"
    """
    suma = np.zeros(len(indices), dtype=nb.complex128)
    for k in range(1, c + 1):
        for i, d in enumerate(indices):
            suma[i] += np.conj(A[k, d]) * B[c - k, d]
    return suma


# @nb.njit("(c16[:])(c16[:, :], c16[:, :], i8)")
@nb.njit(cache=True)
def conv1(A, B, c):
    """
    Performs the convolution of A* and B
    :param A: Coefficients matrix 1 (orders, buses)
    :param B: Coefficients matrix 2 (orders, buses)
    :param c: order of the coefficients
    :param indices: bus indices array
    :return: Array with the convolution for the buses given by "indices"
    """
    suma = np.zeros(A.shape[1], dtype=nb.complex128)
    for k in range(1, c + 1):
        for i in range(A.shape[1]):
            suma[i] += np.conj(A[k, i]) * B[c - k, i]
    return suma


# @nb.njit("(c16[:])(c16[:, :], c16[:, :], i8, i8[:])")
@nb.njit(cache=True)
def conv2(A, B, c, indices):
    """
    Performs the convolution of A and B
    :param A: Coefficients matrix 1 (orders, buses)
    :param B: Coefficients matrix 2 (orders, buses)
    :param c: order of the coefficients
    :param indices: bus indices array
    :return: Array with the convolution for the buses given by "indices"
    """
    suma = np.zeros(len(indices), dtype=nb.complex128)
    for k in range(1, c):
        for i, d in enumerate(indices):
            suma[i] += A[k, d] * B[c - 1 - k, d]
    return suma


# @nb.njit("(c16[:])(c16[:, :], c16[:, :], i8, i8[:])")
@nb.njit(cache=True)
def conv3(A, B, c, indices):
    """
    Performs the convolution of A and B*
    :param A: Coefficients matrix 1 (orders, buses)
    :param B: Coefficients matrix 2 (orders, buses)
    :param c: order of the coefficients
    :param indices: bus indices array
    :return: Array with the convolution for the buses given by "indices"
    """
    suma = np.zeros(len(indices), dtype=nb.complex128)
    for k in range(1, c):
        for i, d in enumerate(indices):
            suma[i] += A[k, d] * np.conj(B[c - k, d])
    return suma


def helm_coefficients_josep(Ybus, Yseries, V0, S0, Ysh0, pq, pv, sl, pqpv, tolerance=1e-6, max_coeff=30, verbose=False,
                            logger: Logger = None):
    """
    Holomorphic Embedding LoadFlow Method as formulated by Josep Fanals Batllori in 2020
    THis function just returns the coefficients for further usage in other routines
    :param Yseries: Admittance matrix of the series elements
    :param V0: vector of specified voltages
    :param S0: vector of specified power
    :param Ysh0: vector of shunt admittances (including the shunts of the branches)
    :param pq: list of pq nodes
    :param pv: list of pv nodes
    :param sl: list of slack nodes
    :param pqpv: sorted list of pq and pv nodes
    :param tolerance: target error (or tolerance)
    :param max_coeff: maximum number of coefficients
    :param verbose: print intermediate information
    :param logger: Logger object to store the debug info
    :return: U, X, Q, V, iterations
    """

    npqpv = len(pqpv)
    npv = len(pv)
    nsl = len(sl)
    n = Yseries.shape[0]

    # --------------------------- PREPARING IMPLEMENTATION -------------------------------------------------------------
    U = np.zeros((max_coeff + 1, npqpv), dtype=complex)  # voltages
    X = np.zeros((max_coeff + 1, npqpv), dtype=complex)  # compute X=1/conj(U)
    Q = np.zeros((max_coeff + 1, npqpv), dtype=complex)  # unknown reactive powers

    if n < 2:
        return U, X, Q, 0

    if verbose:
        logger.add_debug('Yseries', Yseries.toarray())

        df = pd.DataFrame(data=np.c_[Ysh0.imag, S0.real, S0.imag, np.abs(V0)],
                          columns=['Ysh', 'P0', 'Q0', 'V0'])
        logger.add_debug(df.to_string())

    # build the reduced system
    Yred = Yseries[np.ix_(pqpv, pqpv)]  # admittance matrix without slack buses
    Yslack = -Yseries[np.ix_(pqpv, sl)]  # yes, it is the negative of this
    G = Yred.real.copy()  # real parts of Yij
    B = Yred.imag.copy()  # imaginary parts of Yij
    vec_P = S0.real[pqpv]
    vec_Q = S0.imag[pqpv]
    Vslack = V0[sl]
    Ysh = Ysh0[pqpv]
    Vm0 = np.abs(V0[pqpv])
    vec_W = Vm0 * Vm0

    # indices 0 based in the internal scheme
    nsl_counted = np.zeros(n, dtype=int)
    compt = 0
    for i in range(n):
        if i in sl:
            compt += 1
        nsl_counted[i] = compt

    pq_ = pq - nsl_counted[pq]
    pv_ = pv - nsl_counted[pv]
    pqpv_ = np.sort(np.r_[pq_, pv_])

    # .......................CALCULATION OF TERMS [0] ------------------------------------------------------------------

    if nsl > 1:
        U[0, :] = spsolve(Yred, Yslack.sum(axis=1))
    else:
        U[0, :] = spsolve(Yred, Yslack)

    X[0, :] = 1 / np.conj(U[0, :])

    # .......................CALCULATION OF TERMS [1] ------------------------------------------------------------------
    valor = np.zeros(npqpv, dtype=complex)

    # get the current injections that appear due to the slack buses reduction
    I_inj_slack = Yslack[pqpv_, :] * Vslack

    valor[pq_] = I_inj_slack[pq_] - Yslack[pq_].sum(axis=1).A1 + (vec_P[pq_] - vec_Q[pq_] * 1j) * X[0, pq_] - U[0, pq_] * Ysh[pq_]
    valor[pv_] = I_inj_slack[pv_] - Yslack[pv_].sum(axis=1).A1 + (vec_P[pv_]) * X[0, pv_] - U[0, pv_] * Ysh[pv_]

    # compose the right-hand side vector
    RHS = np.r_[valor.real,
                valor.imag,
                vec_W[pv_] - (U[0, pv_] * U[0, pv_]).real  # vec_W[pv_] - 1.0
                ]

    # Form the system matrix (MAT)
    Upv = U[0, pv_]
    Xpv = X[0, pv_]
    VRE = coo_matrix((2 * Upv.real, (np.arange(npv), pv_)), shape=(npv, npqpv)).tocsc()
    VIM = coo_matrix((2 * Upv.imag, (np.arange(npv), pv_)), shape=(npv, npqpv)).tocsc()
    XIM = coo_matrix((-Xpv.imag, (pv_, np.arange(npv))), shape=(npqpv, npv)).tocsc()
    XRE = coo_matrix((Xpv.real, (pv_, np.arange(npv))), shape=(npqpv, npv)).tocsc()
    EMPTY = csc_matrix((npv, npv))

    MAT = vs((hs((G,  -B,   XIM)),
              hs((B,   G,   XRE)),
              hs((VRE, VIM, EMPTY))), format='csc')

    if verbose:
        logger.add_debug("MAT", MAT.toarray())

    # factorize (only once)
    # MAT_LU = factorized(MAT.tocsc())

    # solve
    mat_factorized = factorized(MAT)
    LHS = mat_factorized(RHS)
    # LHS = spsolve(MAT, RHS)

    # update coefficients
    U[1, :] = LHS[:npqpv] + 1j * LHS[npqpv:2 * npqpv]
    Q[0, pv_] = LHS[2 * npqpv:]
    X[1, :] = -X[0, :] * np.conj(U[1, :]) / np.conj(U[0, :])

    # .......................CALCULATION OF TERMS [>=2] ----------------------------------------------------------------
    iter_ = 1
    c = 2
    converged = False
    V = np.empty(n, dtype=complex)
    V[sl] = V0[sl]
    V[pqpv] = U[:c, :].sum(axis=0)
    while c <= max_coeff and not converged:  # c defines the current depth

        valor[pq_] = (vec_P[pq_] - vec_Q[pq_] * 1j) * X[c - 1, pq_] - U[c - 1, pq_] * Ysh[pq_]
        valor[pv_] = -1j * conv2(X, Q, c, pv_) - U[c - 1, pv_] * Ysh[pv_] + X[c - 1, pv_] * vec_P[pv_]

        RHS = np.r_[valor.real,
                    valor.imag,
                    -conv3(U, U, c, pv_).real]

        # LHS = spsolve(MAT, RHS)
        LHS = mat_factorized(RHS)

        # update voltage coefficients
        U[c, :] = LHS[:npqpv] + 1j * LHS[npqpv:2 * npqpv]

        # update reactive power
        Q[c - 1, pv_] = LHS[2 * npqpv:]

        # update voltage inverse coefficients
        X[c, :] = -conv1(U, X, c) / np.conj(U[0, :])

        # compute power mismatch
        V[pqpv] += U[c, :]

        if V.max().real < 10:
            Scalc = compute_power(Ybus, V)
            norm_f = compute_fx_error(compute_fx(Scalc, S0, pqpv, pq))
            converged = (norm_f <= tolerance) and (c % 2)  # we want an odd amount of coefficients
        else:
            # completely erroneous
            break

        iter_ += 1
        c += 1

    return U, X, Q, V, iter_


def helm_josep(Ybus, Yseries, V0, S0, Ysh0, pq, pv, sl, pqpv, tolerance=1e-6, max_coefficients=30, use_pade=True,
               verbose=False, logger: Logger = None) -> NumericPowerFlowResults:
    """
    Holomorphic Embedding LoadFlow Method as formulated by Josep Fanals Batllori in 2020
    :param Ybus: Complete admittance matrix
    :param Yseries: Admittance matrix of the series elements
    :param V0: vector of specified voltages
    :param S0: vector of specified power
    :param Ysh0: vector of shunt admittances (including the shunt "legs" of the pi branches)
    :param pq: list of pq nodes
    :param pv: list of pv nodes
    :param sl: list of slack nodes
    :param pqpv: sorted list of pq and pv nodes
    :param tolerance: target error (or tolerance)
    :param max_coefficients: maximum number of coefficients
    :param use_pade: Use the Padè approximation? otherwise, a simple summation is done
    :param verbose: print intermediate information
    :param logger: Logger object to store the debug info
    :return: V, converged, norm_f, Scalc, iter_, elapsed
    """

    start_time = time.time()

    n = Yseries.shape[0]
    if n < 2:
        return NumericPowerFlowResults(V0, True, 0.0, S0, None, None, None, None, None, None, 0, 0.0)

    # compute the series of coefficients
    U, X, Q, V, iter_ = helm_coefficients_josep(Ybus, Yseries, V0, S0, Ysh0, pq, pv, sl, pqpv,
                                                tolerance=tolerance,
                                                max_coeff=max_coefficients,
                                                verbose=verbose,
                                                logger=logger)

    # --------------------------- RESULTS COMPOSITION ------------------------------------------------------------------
    if verbose:
        logger.add_debug('V coefficients\n', U)

    # compute the final voltage vector
    if use_pade:
        V = V0.copy()
        try:
            V[pqpv] = pade4all(max_coefficients - 1, U, 1)
        except:
            warn('Padè failed :(, using coefficients summation')
            V[pqpv] = U.sum(axis=0)

    # compute power mismatch
    Scalc = compute_power(Ybus, V)
    norm_f = compute_fx_error(compute_fx(Scalc, S0, pqpv, pq))

    # check convergence
    converged = norm_f < tolerance

    elapsed = time.time() - start_time

    return NumericPowerFlowResults(V, converged, norm_f, Scalc, None, None, None, None, None, None, iter_, elapsed)


# if __name__ == '__main__':
#     from GridCal.Engine import FileOpen
#     import pandas as pd
#
#     np.set_printoptions(linewidth=2000, suppress=True)
#     pd.set_option('display.max_rows', 500)
#     pd.set_option('display.max_columns', 500)
#     pd.set_option('display.width', 1000)
#
#     # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE39_1W.gridcal'
#     fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE 14.xlsx'
#     # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/lynn5buspv.xlsx'
#     # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE 118.xlsx'
#     # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/1354 Pegase.xlsx'
#     # fname = 'helm_data1.gridcal'
#
#     grid = FileOpen(fname).open()
#
#     nc = grid.compile_snapshot()
#     inputs = nc.compute()[0]  # pick the first island
#
#     V, converged_, error, Scalc_, iter_, elapsed_ = helm_josep(Ybus=inputs.Ybus,
#                                                                Yseries=inputs.Yseries,
#                                                                V0=inputs.Vbus,
#                                                                S0=inputs.Sbus,
#                                                                Ysh0=inputs.Ysh,
#                                                                pq=inputs.pq,
#                                                                pv=inputs.pv,
#                                                                sl=inputs.ref,
#                                                                pqpv=inputs.pqpv,
#                                                                tolerance=1e-6,
#                                                                max_coefficients=10,
#                                                                use_pade=False,
#                                                                verbose=False)
#     Vm = np.abs(V)
#     Va = np.angle(V)
#     dP = np.abs(inputs.Sbus.real - Scalc_.real)
#     dP[inputs.ref] = 0
#     dQ = np.abs(inputs.Sbus.imag - Scalc_.imag)
#     dQ[inputs.pv] = np.nan
#     dQ[inputs.ref] = np.nan
#     df = pd.DataFrame(data=np.c_[inputs.bus_types, Vm, Va, np.abs(inputs.Vbus), dP, dQ],
#                       columns=['Types', 'Vm', 'Va', 'Vset', 'P mismatch', 'Q mismatch'])
#     print(df)
#     print('Error', error)
#     print('P error', np.max(np.abs(dP)))
#     print('Elapsed', elapsed_)



# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
import os.path

from GridCal.Engine.basic_structures import Logger
from GridCal.Engine.Core.multi_circuit import MultiCircuit
from GridCal.Engine.basic_structures import BranchImpedanceMode
from GridCal.Engine.basic_structures import BusMode
from GridCal.Engine.Devices.enumerations import ConverterControlType, TransformerControlType
from GridCal.Engine.Devices import *
from GridCal.Engine.basic_structures import Logger, SolverType, ReactivePowerControlMode, TapsControlMode
from GridCal.Engine.Simulations.PowerFlow.power_flow_options import PowerFlowOptions
from GridCal.Engine.Simulations.PowerFlow.power_flow_results import PowerFlowResults
from GridCal.Engine.IO.file_system import get_create_gridcal_folder
import GridCal.Engine.basic_structures as bs

try:
    import bentayga as btg

    # activate
    if not btg.is_license_activated():
        btg_license = os.path.join(get_create_gridcal_folder(), 'bentayga.lic')
        if os.path.exists(btg_license):

            try:
                btg.activate_license(btg_license)
                if btg.is_license_activated():
                    BENTAYGA_AVAILABLE = True
                else:
                    print('Bentayga v' + btg.get_version(),
                          "installed, tried to activate with {} but the license did not work :/".format(btg_license))
                    BENTAYGA_AVAILABLE = False
            except RuntimeError:
                print("Bentayga: Error reading the license file :(")
                BENTAYGA_AVAILABLE = False
        else:
            print('Bentayga v' + btg.get_version(), "installed but not licensed")
            BENTAYGA_AVAILABLE = False
    else:
        print('Bentayga v' + btg.get_version())
        BENTAYGA_AVAILABLE = True

except ImportError:
    BENTAYGA_AVAILABLE = False
    print('Bentayga is not available')

# numpy integer type for bentayga's uword
BINT = np.ulonglong


def add_btg_buses(circuit: MultiCircuit, btg_circuit: "btg.Circuit", time_series: bool, ntime: int=1):
    """
    Convert the buses to bentayga buses
    :param circuit: GridCal circuit
    :param btg_circuit: bentayga circuit
    :param time_series: compile the time series from GridCal? otherwise, just the snapshot
    :param ntime: number of time steps
    :return: bus dictionary buses[uuid] -> Bus
    """
    areas_dict = {elm: k for k, elm in enumerate(circuit.areas)}
    bus_dict = dict()

    for i, bus in enumerate(circuit.buses):

        elm = btg.Node(uuid=bus.idtag,
                       name=bus.name,
                       time_steps=ntime,
                       is_slack=bus.is_slack,
                       is_dc=bus.is_dc,
                       nominal_voltage=bus.Vnom)

        if time_series and ntime > 1:
            elm.active = bus.active_prof.astype(BINT)
        else:
            elm.active = np.ones(ntime, dtype=BINT) * int(bus.active)

        btg_circuit.add_node(elm)
        bus_dict[elm.uuid] = elm

    return bus_dict


def add_btg_loads(circuit: MultiCircuit, btg_circuit: "btg.Circuit", bus_dict, time_series: bool, ntime=1):
    """

    :param circuit: GridCal circuit
    :param btg_circuit: bentayga circuit
    :param time_series: compile the time series from GridCal? otherwise just the snapshot
    :param bus_dict: dictionary of bus id to bentayga bus object
    :param ntime: number of time steps
    """

    devices = circuit.get_loads()
    for k, elm in enumerate(devices):

        load = btg.Load(uuid=elm.idtag,
                        name=elm.name,
                        bus=bus_dict[elm.bus.idtag],
                        time_steps=ntime,
                        P0=elm.P,
                        Q0=elm.Q)

        if time_series:
            load.active = elm.active_prof.astype(BINT)
            load.P = elm.P_prof
            load.Q = elm.Q_prof
        else:
            load.active = np.ones(ntime, dtype=BINT) * int(elm.active)

        btg_circuit.add_load(load)


def add_btg_static_generators(circuit: MultiCircuit, btg_circuit: "btg.Circuit", bus_dict, time_series: bool, ntime=1):
    """

    :param circuit: GridCal circuit
    :param btg_circuit: bentayga circuit
    :param time_series: compile the time series from GridCal? otherwise just the snapshot
    :param bus_dict: dictionary of bus id to bentayga bus object
    :param ntime: number of time steps
    """
    devices = circuit.get_static_generators()
    for k, elm in enumerate(devices):

        load = btg.Load(uuid=elm.idtag,
                        name=elm.name,
                        bus=bus_dict[elm.bus.idtag],
                        time_steps=ntime,
                        P0=-elm.P,
                        Q0=-elm.Q)

        if time_series:
            load.active = elm.active_prof.astype(BINT)
            load.P = -elm.P_prof
            load.Q = -elm.Q_prof
        else:
            load.active = np.ones(ntime, dtype=BINT) * int(elm.active)

        btg_circuit.add_load(load)


def add_btg_shunts(circuit: MultiCircuit, btg_circuit: "btg.Circuit", bus_dict, time_series: bool, ntime=1):
    """

    :param circuit: GridCal circuit
    :param btg_circuit: bentayga circuit
    :param time_series: compile the time series from GridCal? otherwise just the snapshot
    :param bus_dict: dictionary of bus id to bentayga bus object
    :param ntime: number of time steps
    """
    devices = circuit.get_shunts()
    for k, elm in enumerate(devices):

        sh = btg.ShuntFixed(uuid=elm.idtag,
                            name=elm.name,
                            bus=bus_dict[elm.bus.idtag],
                            time_steps=ntime,
                            G0=elm.G,
                            B0=elm.B)

        if time_series:
            sh.active = elm.active_prof.astype(BINT)
            sh.G = elm.G_prof
            sh.B = elm.B_prof
        else:
            sh.active = np.ones(ntime, dtype=BINT) * int(elm.active)

        btg_circuit.add_shunt_fixed(sh)


def add_btg_generators(circuit: MultiCircuit, btg_circuit: "btg.Circuit", bus_dict, time_series: bool, ntime=1):
    """

    :param circuit: GridCal circuit
    :param btg_circuit: bentayga circuit
    :param time_series: compile the time series from GridCal? otherwise just the snapshot
    :param bus_dict: dictionary of bus id to bentayga bus object
    :param ntime: number of time steps
    """
    devices = circuit.get_generators()

    for k, elm in enumerate(devices):

        gen = btg.Generator(uuid=elm.idtag,
                            name=elm.name,
                            bus=bus_dict[elm.bus.idtag],
                            time_steps=ntime,
                            P0=elm.P,
                            Q0=0,
                            Vset0=elm.Vset)
        gen.Qmin = elm.Qmin
        gen.Qmax = elm.Qmax
        gen.Pmin = elm.Pmin
        gen.Pmax = elm.Pmax
        gen.generation_cost = elm.Cost

        if time_series:
            gen.active = elm.active_prof.astype(BINT)
            gen.P = elm.P_prof
            gen.vset = elm.Vset_prof
        else:
            gen.active = np.ones(ntime, dtype=BINT) * int(elm.active)
            gen.P = np.ones(ntime, dtype=float) * elm.P
            gen.vset = np.ones(ntime, dtype=float) * elm.Vset

        btg_circuit.add_generator(gen)


def get_battery_data(circuit: MultiCircuit, btg_circuit: "btg.Circuit", bus_dict, time_series: bool, ntime=1):
    """

    :param circuit: GridCal circuit
    :param btg_circuit: bentayga circuit
    :param time_series: compile the time series from GridCal? otherwise just the snapshot
    :param bus_dict: dictionary of bus id to bentayga bus object
    :param ntime: number of time steps
    """
    devices = circuit.get_batteries()

    for k, elm in enumerate(devices):

        gen = btg.Battery(uuid=elm.idtag,
                          name=elm.name,
                          bus=bus_dict[elm.bus.idtag],
                          time_steps=ntime,
                          capacity=elm.Enom,
                          P0=elm.P,
                          Q0=0,
                          Vset0=elm.Vset)

        gen.soc_max = elm.max_soc
        gen.soc_min = elm.min_soc
        gen.charge_efficiency = elm.charge_efficiency
        gen.discharge_efficiency = elm.discharge_efficiency
        gen.generation_cost = elm.Cost
        gen.Qmin = elm.Qmin
        gen.Qmax = elm.Qmax
        gen.Pmin = elm.Pmin
        gen.Pmax = elm.Pmax

        if time_series:
            gen.active = elm.active_prof.astype(BINT)
            gen.P = elm.P_prof
            gen.vset = elm.Vset_prof
        else:
            gen.active = np.ones(ntime, dtype=BINT) * int(elm.active)
            gen.P = np.ones(ntime, dtype=float) * elm.P
            gen.vset = np.ones(ntime, dtype=float) * elm.Vset

        btg_circuit.add_battery(gen)


def add_btg_line(circuit: MultiCircuit, btg_circuit: "btg.Circuit", bus_dict, time_series: bool, ntime=1):
    """

    :param circuit: GridCal circuit
    :param btg_circuit: bentayga circuit
    :param time_series: compile the time series from GridCal? otherwise just the snapshot
    :param bus_dict: dictionary of bus id to bentayga bus object
    :param ntime: number of time steps
    """

    # Compile the lines
    for i, elm in enumerate(circuit.lines):
        lne = btg.AcLine(uuid=elm.idtag,
                         name=elm.name,
                         node_from=bus_dict[elm.bus_from.idtag],
                         node_to=bus_dict[elm.bus_to.idtag],
                         time_steps=ntime,
                         length=elm.length,
                         rate=elm.rate,
                         active_default=elm.active,
                         r1=elm.R,
                         x1=elm.X,
                         b1=elm.B)

        lne.monitor_loading = np.ones(ntime, dtype=BINT) * int(elm.monitor_loading)
        lne.contingency_enabled = np.ones(ntime, dtype=BINT) * int(elm.contingency_enabled)

        if time_series:
            lne.active = elm.active_prof.astype(BINT)
            lne.rates = elm.rate_prof
            lne.contingency_rates = elm.rate_prof * elm.contingency_factor

        btg_circuit.add_ac_line(lne)


def get_transformer_data(circuit: MultiCircuit, btg_circuit: "btg.Circuit", bus_dict, time_series: bool, ntime=1):
    """

    :param circuit: GridCal circuit
    :param btg_circuit: bentayga circuit
    :param time_series: compile the time series from GridCal? otherwise just the snapshot
    :param bus_dict: dictionary of bus id to bentayga bus object
    :param ntime: number of time steps
    """
    for i, elm in enumerate(circuit.transformers2w):
        tr2 = btg.Transformer2WAll(uuid=elm.idtag,
                                   name=elm.name,
                                   node_from=bus_dict[elm.bus_from.idtag],
                                   node_to=bus_dict[elm.bus_to.idtag],
                                   time_steps=ntime,
                                   HV=elm.HV,
                                   LV=elm.LV,
                                   rate=elm.rate,
                                   active_default=elm.active,
                                   r1=elm.R,
                                   x1=elm.X,
                                   g1=elm.G,
                                   b1=elm.B)

        tr2.monitor_loading = np.ones(ntime, dtype=BINT) * int(elm.monitor_loading)
        tr2.contingency_enabled = np.ones(ntime, dtype=BINT) * int(elm.contingency_enabled)

        if time_series:
            tr2.active = elm.active_prof.astype(BINT)
            tr2.rates = elm.rate_prof
            tr2.contingency_rates = elm.rate_prof * elm.contingency_factor
            tr2.tap = elm.tap_module_prof
            tr2.phase = elm.angle_prof
        else:
            tr2.tap = np.ones(ntime, dtype=float) * elm.tap_module
            tr2.phase = np.ones(ntime, dtype=float) * elm.angle

        btg_circuit.add_transformer_all(tr2)


def get_vsc_data(circuit: MultiCircuit, btg_circuit: "btg.Circuit", bus_dict, time_series: bool, ntime=1):
    """

    :param circuit: GridCal circuit
    :param btg_circuit: bentayga circuit
    :param time_series: compile the time series from GridCal? otherwise just the snapshot
    :param bus_dict: dictionary of bus id to bentayga bus object
    :param ntime: number of time steps
    """
    for i, elm in enumerate(circuit.vsc_devices):
        vsc = btg.VSC(uuid=elm.idtag,
                      name=elm.name,
                      node_from=bus_dict[elm.bus_from.idtag],
                      node_to=bus_dict[elm.bus_to.idtag],
                      time_steps=ntime,
                      rate=elm.rate,
                      active_default=elm.active,
                      r1=elm.R1,
                      x1=elm.X1,
                      g0=elm.G0,
                      beq=elm.Beq,
                      beq_max=elm.Beq_max,
                      beq_min=elm.Beq_min,
                      k=elm.k,
                      tap=elm.m,
                      tap_max=elm.m_max,
                      tap_min=elm.m_min,
                      phase=elm.theta,
                      phase_max=elm.theta_max,
                      phase_min=elm.theta_min,
                      Pf_set=elm.Pdc_set,
                      vac_set=elm.Vac_set,
                      vdc_set=elm.Vdc_set,
                      kdp=elm.kdp,
                      alpha1=elm.alpha1,
                      alpha2=elm.alpha2,
                      alpha3=elm.alpha3)

        vsc.monitor_loading = np.ones(ntime, dtype=BINT) * int(elm.monitor_loading)
        vsc.contingency_enabled = np.ones(ntime, dtype=BINT) * int(elm.contingency_enabled)

        if time_series:
            vsc.active = elm.active_prof.astype(BINT)
            vsc.rates = elm.rate_prof
            vsc.contingency_rates = elm.rate_prof * elm.contingency_factor

        btg_circuit.add_vsc(vsc)


def get_dc_line_data(circuit: MultiCircuit, btg_circuit: "btg.Circuit", bus_dict, time_series: bool, ntime=1):
    """

    :param circuit: GridCal circuit
    :param btg_circuit: bentayga circuit
    :param time_series: compile the time series from GridCal? otherwise just the snapshot
    :param bus_dict: dictionary of bus id to bentayga bus object
    :param ntime: number of time steps
    """
    # Compile the lines
    for i, elm in enumerate(circuit.dc_lines):
        lne = btg.DcLine(uuid=elm.idtag,
                         name=elm.name,
                         node_from=bus_dict[elm.bus_from.idtag],
                         node_to=bus_dict[elm.bus_to.idtag],
                         time_steps=ntime,
                         length=elm.length,
                         rate=elm.rate,
                         active_default=elm.active,
                         r1=elm.R)

        lne.monitor_loading = np.ones(ntime, dtype=BINT) * int(elm.monitor_loading)
        lne.contingency_enabled = np.ones(ntime, dtype=BINT) * int(elm.contingency_enabled)

        if time_series:
            lne.active = elm.active_prof.astype(BINT)
            lne.rates = elm.rate_prof
            lne.contingency_rates = elm.rate_prof * elm.contingency_factor

        btg_circuit.add_dc_line(lne)


def get_hvdc_data(circuit: MultiCircuit, btg_circuit: "btg.Circuit", bus_dict, time_series: bool, ntime=1):
    """

    :param circuit: GridCal circuit
    :param btg_circuit: bentayga circuit
    :param time_series: compile the time series from GridCal? otherwise just the snapshot
    :param bus_dict: dictionary of bus id to bentayga bus object
    :param ntime: number of time steps
    """

    cmode_dict = {HvdcControlType.type_0_free: btg.HvdcControlType.free,
                  HvdcControlType.type_1_Pset: btg.HvdcControlType.Pdc}

    for i, elm in enumerate(circuit.hvdc_lines):
        hvdc = btg.HvdcLine(uuid=elm.idtag,
                            name=elm.name,
                            node_from=bus_dict[elm.bus_from.idtag],
                            node_to=bus_dict[elm.bus_to.idtag],
                            time_steps=ntime,
                            length=elm.length,
                            rate=elm.rate,
                            active_default=elm.active,
                            r1=elm.r,
                            Pset=elm.Pset,
                            v_set_f=elm.Vset_f,
                            v_set_t=elm.Vset_t,
                            min_firing_angle_f=elm.min_firing_angle_f,
                            min_firing_angle_t=elm.min_firing_angle_t,
                            max_firing_angle_f=elm.max_firing_angle_f,
                            max_firing_angle_t=elm.max_firing_angle_t,
                            control_mode=cmode_dict[elm.control_mode])

        # hvdc.monitor_loading = elm.monitor_loading
        # hvdc.contingency_enabled = elm.contingency_enabled

        if time_series:
            hvdc.active = elm.active_prof.astype(BINT)
            hvdc.rates = elm.rate_prof
            hvdc.v_set_f = elm.Vset_f_prof
            hvdc.v_set_t = elm.Vset_t_prof
            hvdc.contingency_rates = elm.rate_prof * elm.contingency_factor
            hvdc.angle_droop = elm.angle_droop_prof
        else:
            hvdc.contingency_rates = elm.rate * elm.contingency_factor
            hvdc.angle_droop = elm.angle_droop

        btg_circuit.add_hvdc_line(hvdc)


def to_bentayga(circuit: MultiCircuit, time_series: bool):
    """
    Convert GridCal circuit to Bentayga
    :param circuit: MultiCircuit
    :param time_series: compile the time series from GridCal? otherwise just the snapshot
    :return: btg.Circuit instance
    """
    ntime = circuit.get_time_number() if time_series else 1
    if ntime == 0:
        ntime = 1

    btgCircuit = btg.Circuit(uuid=circuit.idtag, name=circuit.name, time_steps=ntime)

    bus_dict = add_btg_buses(circuit, btgCircuit, time_series, ntime)
    add_btg_loads(circuit, btgCircuit, bus_dict, time_series, ntime)
    add_btg_static_generators(circuit, btgCircuit, bus_dict, time_series, ntime)
    add_btg_shunts(circuit, btgCircuit, bus_dict, time_series, ntime)
    add_btg_generators(circuit, btgCircuit, bus_dict, time_series, ntime)
    get_battery_data(circuit, btgCircuit, bus_dict, time_series, ntime)
    add_btg_line(circuit, btgCircuit, bus_dict, time_series, ntime)
    get_transformer_data(circuit, btgCircuit, bus_dict, time_series, ntime)
    get_vsc_data(circuit, btgCircuit, bus_dict, time_series, ntime)
    get_dc_line_data(circuit, btgCircuit, bus_dict, time_series, ntime)
    get_hvdc_data(circuit, btgCircuit, bus_dict, time_series, ntime)

    return btgCircuit


class FakeAdmittances:

    def __init__(self):
        self.Ybus = None
        self.Yf = None
        self.Yt = None


def get_snapshots_from_bentayga(circuit: MultiCircuit):

    from GridCal.Engine.Core.snapshot_pf_data import SnapshotData

    btgCircuit = to_bentayga(circuit, time_series=False)

    btg_data_lst = btg.compile_at(btgCircuit, t=0)

    data_lst = list()

    for btg_data in btg_data_lst:

        data = SnapshotData(nbus=0,
                            nline=0,
                            ndcline=0,
                            ntr=0,
                            nvsc=0,
                            nupfc=0,
                            nhvdc=0,
                            nload=0,
                            ngen=0,
                            nbatt=0,
                            nshunt=0,
                            nstagen=0,
                            sbase=0,
                            ntime=1)

        data.Vbus_ = btg_data.Vbus.reshape(-1, 1)
        data.Sbus_ = btg_data.Sbus.reshape(-1, 1)
        data.Ibus_ = btg_data.Ibus
        data.branch_data.branch_names = np.array(btg_data.branch_data.names)
        data.branch_data.tap_f = btg_data.branch_data.virtual_tap_f
        data.branch_data.tap_t = btg_data.branch_data.virtual_tap_t

        data.bus_data.bus_names = np.array(btg_data.bus_data.names)

        data.Admittances = FakeAdmittances()
        data.Admittances.Ybus = btg_data.admittances.Ybus
        data.Admittances.Yf = btg_data.admittances.Yf
        data.Admittances.Yt = btg_data.admittances.Yt

        data.Bbus_ = btg_data.linear_admittances.Bbus
        data.Bf_ = btg_data.linear_admittances.Bf

        data.Yseries_ = btg_data.split_admittances.Yseries
        data.Yshunt_ = btg_data.split_admittances.Yshunt

        data.B1_ = btg_data.fast_decoupled_admittances.B1
        data.B2_ = btg_data.fast_decoupled_admittances.B2

        data.Cf_ = btg_data.Cf
        data.Ct_ = btg_data.Ct

        data.bus_data.bus_types = [x.value for x in btg_data.bus_data.bus_types]
        data.pq_ = btg_data.bus_types_data.pq
        data.pv_ = btg_data.bus_types_data.pv
        data.vd_ = btg_data.bus_types_data.vd
        data.pqpv_ = btg_data.bus_types_data.pqpv

        data.original_bus_idx = btg_data.bus_data.original_indices
        data.original_branch_idx = btg_data.branch_data.original_indices

        data.Qmax_bus_ = btg_data.Qmax_bus
        data.Qmin_bus_ = btg_data.Qmin_bus

        data.iPfsh = btg_data.control_indices.iPfsh
        data.iQfma = btg_data.control_indices.iQfma
        data.iBeqz = btg_data.control_indices.iBeqz
        data.iBeqv = btg_data.control_indices.iBeqv
        data.iVtma = btg_data.control_indices.iVtma
        data.iQtma = btg_data.control_indices.iQtma
        data.iPfdp = btg_data.control_indices.iPfdp
        data.iVscL = btg_data.control_indices.iVscL
        data.VfBeqbus = btg_data.control_indices.iVfBeqBus
        data.Vtmabus = btg_data.control_indices.iVtmaBus

        data_lst.append(data)

    return data_lst


def get_bentayga_pf_options(opt: PowerFlowOptions):
    """
    Translate GridCal power flow options to Bentayga power flow options
    :param opt:
    :return:
    """
    solver_dict = {SolverType.NR: btg.PowerFlowSolvers.NewtonRaphson,
                   SolverType.DC: btg.PowerFlowSolvers.LinearDc,
                   # SolverType.HELM: nn.NativeSolverType.HELM,
                   # SolverType.IWAMOTO: nn.NativeSolverType.IWAMOTO,
                   SolverType.LM: btg.PowerFlowSolvers.LevenbergMarquardt,
                   # SolverType.LACPF: nn.NativeSolverType.LACPF,
                   SolverType.FASTDECOUPLED: btg.PowerFlowSolvers.FastDecoupled
                   }

    q_control_dict = {ReactivePowerControlMode.NoControl: btg.QControlMode.NoControl,
                      ReactivePowerControlMode.Direct: btg.QControlMode.Direct}

    if opt.solver_type in solver_dict.keys():
        solver_type = solver_dict[opt.solver_type]
    else:
        solver_type = btg.PowerFlowSolvers.NewtonRaphson

    return btg.PowerFlowOptions(solver=solver_type,
                                tolerance=opt.tolerance,
                                max_iter=opt.max_iter,
                                retry_with_other_methods=opt.retry_with_other_methods,
                                q_control_mode=q_control_dict[opt.control_Q])


def bentayga_pf(circuit: MultiCircuit, opt: PowerFlowOptions, time_series=False):
    """
    Bentayga power flow
    :param circuit: MultiCircuit instance
    :param opt: Power Flow Options
    :param time_series: Compile with GridCal time series?
    :return: Bentayga Power flow results object
    """
    btgCircuit = to_bentayga(circuit, time_series=time_series)

    pf_options = get_bentayga_pf_options(opt)
    logger = btg.Logger()
    pf_res = btg.run_power_flow(circuit=btgCircuit,
                                options=pf_options,
                                logger=logger,
                                parallel=True)

    return pf_res


def bentayga_linear_matrices(circuit: MultiCircuit, distributed_slack=False):
    """
    Bentayga linear analysis
    :param circuit: MultiCircuit instance
    :param distributed_slack: distribute the PTDF slack
    :return: Bentayga LinearAnalysisMatrices object
    """
    btg_circuit = to_bentayga(circuit, time_series=False)
    lin_mat = btg.compute_linear_matrices_at(t=0,
                                             circuit=btg_circuit,
                                             distributed_slack=distributed_slack)

    return lin_mat


def translate_bentayga_pf_results(grid: MultiCircuit, res) -> PowerFlowResults:
    results = PowerFlowResults(n=grid.get_bus_number(),
                               m=grid.get_branch_number_wo_hvdc(),
                               n_tr=grid.get_transformers2w_number(),
                               n_hvdc=grid.get_hvdc_number(),
                               bus_names=res.bus_names,
                               branch_names=res.branch_names,
                               transformer_names=[],
                               hvdc_names=res.hvdc_names,
                               bus_types=res.bus_types)

    results.voltage = res.V[0, :]
    results.Sbus = res.S[0, :]
    results.Sf = res.Sf[0, :]
    results.St = res.St[0, :]
    results.loading = res.loading[0, :]
    results.losses = res.losses[0, :]
    results.Vbranch = res.Vbranch[0, :]
    results.If = res.If[0, :]
    results.It = res.It[0, :]
    results.Beq = res.Beq[0, :]
    results.m = res.tap_modules[0, :]
    results.theta = res.tap_angles[0, :]
    results.F = res.F
    results.T = res.T
    results.hvdc_F = res.F_hvdc
    results.hvdc_T = res.T_hvdc
    results.hvdc_Pf = res.hvdc_Pf[0, :]
    results.hvdc_Pt = res.hvdc_Pt[0, :]
    results.hvdc_loading = res.hvdc_loading[0, :]
    results.hvdc_losses = res.hvdc_losses[0, :]
    results.bus_area_indices = grid.get_bus_area_indices()
    results.area_names = [a.name for a in grid.areas]

    for rep in res.stats[0]:
        report = bs.ConvergenceReport()
        for i in range(len(rep.converged)):
            report.add(method=rep.solver[i].name,
                       converged=rep.converged[i],
                       error=rep.norm_f[i],
                       elapsed=rep.elapsed[i],
                       iterations=rep.iterations[i])
            results.convergence_reports.append(report)

    return results

def debug_bentayga_circuit_at(btg_circuit: "btg.Circuit", t: int = None):

    if t is None:
        t = 0

    data = btg.compile_at(btg_circuit, t=t)

    for i in range(len(data)):

        print('_' * 200)
        print('Island', i)
        print('_' * 200)

        print("Ybus")
        print(data[i].admittances.Ybus.toarray())

        print('Yseries')
        print(data[i].split_admittances.Yseries.toarray())

        print('Yshunt')
        print(data[i].split_admittances.Yshunt)

        print("Bbus")
        print(data[i].linear_admittances.Bbus.toarray())

        print('B1')
        print(data[i].fast_decoupled_admittances.B1.toarray())

        print('B2')
        print(data[i].fast_decoupled_admittances.B2.toarray())

        print('Sbus')
        print(data[i].Sbus)

        print('Vbus')
        print(data[i].Vbus)

        print('Qmin')
        print(data[i].Qmin_bus)

        print('Qmax')
        print(data[i].Qmax_bus)



# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
import os
import chardet
import pandas as pd
from math import sqrt
from typing import Set, Dict, List, Tuple


def rfid2uuid(val):
    return val.replace('-', '').replace('_', '')


def index_find(string, start, end):
    """
    version of substring that matches
    :param string: string
    :param start: string to start splitting
    :param end: string to end splitting
    :return: string between start and end
    """
    return string.partition(start)[2].partition(end)[0]


def str2num(val: str):
    """
    Try to convert to number, else keep as string
    :param val: String value
    :return: int, float or string
    """
    try:
        return int(val)
    except ValueError:
        try:
            return float(val)
        except ValueError:
            return val


class GeneralContainer:

    def __init__(self, rfid, tpe, resources=list(), class_replacements=dict()):
        """
        General CIM object container
        :param rfid: RFID
        :param tpe: type of the object (class)
        """

        # dictionary of properties read from the xml
        self.properties = dict()

        # dictionary of objects that reference this object
        self.references_to_me = dict()

        # store the object type
        self.tpe = tpe

        # pick the object id
        self.rfid = rfid
        self.uuid = rfid2uuid(rfid)

        self.class_replacements = class_replacements
        self.resources = resources

        self.name = ''

    def __repr__(self):
        return self.rfid

    def __hash__(self):
        # alternatively, return hash(repr(self))
        return int(self.uuid, 16)  # hex string to int

    def __lt__(self, other):
        return self.__hash__() < other.__hash__()

    def __eq__(self, other):
        return self.rfid == other.rfid

    def add_reference(self, obj: "GeneralContainer"):
        """
        Adds a categorized reference to this object
        :param obj:
        :return:
        """
        if obj.tpe in self.references_to_me.keys():
            self.references_to_me[obj.tpe].add(obj)
        else:
            self.references_to_me[obj.tpe] = {obj}

    def parse_line(self, xml_line):
        """
        Parse xml line that eligibly belongs to this object
        :param xml_line: xml text line
        """

        # the parsers are lists of 2 sets of separators
        # the first separator tries to substring the property name
        # the second tries to substring the property value
        parsers = [[('.', '>'), ('>', '<')],
                   [('.', ' rdf:resource'), ('rdf:resource="', '"')]]

        for L1, L2 in parsers:
            # try to parse the property
            prop = index_find(xml_line, L1[0], L1[1]).strip()

            # try to parse the value
            val = index_find(xml_line, L2[0], L2[1])

            # remove the pound
            if len(val) > 0:
                if val[0] == '#':
                    val = val[1:]

            val = val.replace('\n', '')

            if prop != "" and val != "":
                self.properties[prop] = val

                if hasattr(self, prop):
                    setattr(self, prop, str2num(val))

    def merge(self, other):
        """
        Merge the properties of this object with another
        :param other: GeneralContainer instance
        """
        self.properties = {**self.properties, **other.properties}

    def print(self):
        print('Type:' + self.tpe)
        print('Id:' + self.rfid)

        for key in self.properties.keys():
            val = self.properties[key]

            if type(val) == GeneralContainer:
                for key2 in val.properties.keys():
                    val2 = val.properties[key2]
                    print(key, '->', key2, ':', val2)
            else:
                print(key, ':', val)

    def __str__(self):
        return self.tpe + ':' + self.rfid

    def get_xml(self, level=0):

        """
        Returns an XML representation of the object
        Args:
            level:

        Returns:

        """

        """
        <cim:IEC61970CIMVersion rdf:ID="version">
            <cim:IEC61970CIMVersion.version>IEC61970CIM16v29a</cim:IEC61970CIMVersion.version>
            <cim:IEC61970CIMVersion.date>2015-07-15</cim:IEC61970CIMVersion.date>
        </cim:IEC61970CIMVersion>
        """

        l1 = '  ' * level  # start/end tabbing
        l2 = '  ' * (level + 1)  # middle tabbing

        # header
        xml = l1 + '<cim:' + self.tpe + ' rdf:ID="' + self.rfid + '">\n'

        # properties
        for prop, value in self.properties.items():
            v = str(value).replace(' ', '_')

            # eventually replace the class of the property, because CIM is so well designed...
            if prop in self.class_replacements.keys():
                cls = self.class_replacements[prop]
            else:
                cls = self.tpe

            if prop in self.resources:
                xml += l2 + '<cim:' + cls + '.' + prop + ' rdf:resource="#' + v + '" />\n'
            else:
                xml += l2 + '<cim:' + cls + '.' + prop + '>' + v + '</cim:' + cls + '.' + prop + '>\n'

        # closing
        xml += l1 + '</cim:' + self.tpe + '>\n'

        return xml

    def get_dict(self):
        """
        Get dictionary with the data
        :return: Dictionary
        """
        return {'rfid': self.rfid,
                'uuid': self.uuid,
                'name': self.name}


class MonoPole(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)

    def get_topological_node(self):
        """
        Get the TopologyNodes of this branch
        :return: two TopologyNodes or nothing
        """
        try:
            terminals = list(self.references_to_me['Terminal'])

            if len(terminals) == 1:
                n1 = terminals[0].TopologicalNode
                return n1
            else:
                return None

        except KeyError:
            return None

    def get_bus(self):
        """
        Get the associated bus
        :return:
        """
        tp = self.get_topological_node()
        if tp is None:
            return None
        else:
            return tp.get_bus()

    def get_dict(self):
        """
        Get dictionary with the data
        :return: Dictionary
        """
        tp = self.get_topological_node()
        bus = tp.get_bus() if tp is not None else None

        d = super().get_dict()
        d['TopologicalNode'] = '' if tp is None else tp.uuid
        d['BusbarSection'] = '' if bus is None else bus.uuid
        return d


class DiPole(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)

    def get_topological_nodes(self) -> Tuple["TopologicalNode", "TopologicalNode"]:
        """
        Get the TopologyNodes of this branch
        :return: (TopologyNodes, TopologyNodes) or (None, None)
        """
        try:
            terminals = list(self.references_to_me['Terminal'])

            if len(terminals) == 2:
                n1 = terminals[0].TopologicalNode
                n2 = terminals[1].TopologicalNode
                return n1, n2
            else:
                return None, None

        except KeyError:
            return None, None

    def get_buses(self) -> Tuple["BusbarSection", "BusbarSection"]:
        """
        Get the associated bus
        :return: (BusbarSection, BusbarSection) or (None, None)
        """
        t1, t2 = self.get_topological_nodes()
        b1 = t1.get_bus() if t1 is not None else None
        b2 = t2.get_bus() if t2 is not None else None
        return b1, b2

    def get_dict(self):
        """
        Get dictionary with the data
        :return: Dictionary
        """
        t1, t2 = self.get_topological_nodes()
        b1 = t1.get_bus() if t1 is not None else None
        b2 = t2.get_bus() if t2 is not None else None

        d = super().get_dict()
        d['TopologicalNode1'] = '' if t1 is None else t1.uuid
        d['TopologicalNode2'] = '' if t2 is None else t2.uuid
        d['BusbarSection1'] = '' if b1 is None else b1.uuid
        d['BusbarSection2'] = '' if b2 is None else b2.uuid
        return d


class BaseVoltage(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)

        self.nominalVoltage = 0


class Terminal(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)

        self.TopologicalNode = None
        self.ConductingEquipment = None  # pointer to the Bus
        self.name = ''
        self.connected = True

    def get_voltage(self):
        """
        Get the voltage of this terminal
        :return: Voltage or None
        """
        if self.TopologicalNode is not None:
            return self.TopologicalNode.get_voltage()
        else:
            return None


class ConnectivityNode(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)

        self.TopologicalNode = None
        self.ConnectivityNodeContainer = None
        self.name = ''
        self.shortName = ''
        self.description = ''
        self.fromEndName = ''
        self.fromEndNameTSO = ''
        self.fromEndIsoCode = ''
        self.toEndName = ''
        self.toEndNameTSO = ''
        self.toEndIsoCode = ''
        self.boundaryPoint = ''


class TopologicalNode(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)

        self.BaseVoltage = None
        self.ConnectivityNodeContainer = None

    def get_voltage(self):

        if self.BaseVoltage is not None:
            return self.BaseVoltage.nominalVoltage
        else:
            return None

    def get_bus(self):
        """
        Get an associated BusBar, if any
        :return: BusbarSection or None is not fond
        """
        try:
            terms = self.references_to_me['Terminal']
            for term in terms:
                if isinstance(term.ConductingEquipment, BusbarSection):
                    return term.ConductingEquipment

        except KeyError:
            return None

    def get_dict(self):

        d = super().get_dict()

        v = self.get_voltage()
        d['Vnom'] = v if v is not None else ''

        return d


class BusbarSection(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)

        self.EquipmentContainer = None

    def get_topological_nodes(self):
        """
        Get the associated TopologicalNode instances
        :return: list of TopologicalNode instances
        """
        try:
            terms = self.references_to_me['Terminal']
            return [term.TopologicalNode for term in terms]
        except KeyError:
            return list()

    def get_topological_node(self):
        """
        Get the first TopologicalNode found
        :return: first TopologicalNode found
        """
        try:
            terms = self.references_to_me['Terminal']
            for term in terms:
                return term.TopologicalNode
        except KeyError:
            return list()

    def get_dict(self):

        d = super().get_dict()

        voltage_level = self.EquipmentContainer

        if voltage_level is not None:
            d['Vnom'] = voltage_level.BaseVoltage.nominalVoltage
            d['VoltageLevel'] = voltage_level.uuid
            substation = voltage_level.Substation

            if substation is not None:
                d['Substation'] = substation.uuid
                province = substation.Region

                if province is not None:
                    d['SubGeographicalRegion'] = province.uuid
                    country = province.Region

                    if country is not None:
                        d['GeographicalRegion'] = country.uuid
                    else:
                        d['GeographicalRegion'] = ''

                else:
                    d['SubGeographicalRegion'] = ''
                    d['GeographicalRegion'] = ''

            else:
                d['Substation'] = ''
                d['SubGeographicalRegion'] = ''
                d['GeographicalRegion'] = ''

        else:
            d['VoltageLevel'] = ''
            d['Substation'] = ''
            d['SubGeographicalRegion'] = ''
            d['GeographicalRegion'] = ''

        return d


class Substation(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)

        self.name = ''
        self.Region = None


class OperationalLimitSet(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)

        self.name = ''
        self.Terminal = None


class OperationalLimitType(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)

        self.name = ''
        self.limitType = ''
        self.direction = ''


class GeographicalRegion(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)


class SubGeographicalRegion(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)

        self.Region = None


class VoltageLevel(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)

        self.Substation = None
        self.BaseVoltage = None
        self.highVoltageLimit = 0
        self.lowVoltageLimit = 0


class VoltageLimit(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)

        self.OperationalLimitType = None
        self.OperationalLimitSet = None
        self.value = 0


class CurrentLimit(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)

        self.OperationalLimitType = None
        self.OperationalLimitSet = None
        self.value = 0


class EquivalentNetwork(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)


class EquivalentInjection(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)

        self.regulationCapability = False
        self.BaseVoltage = None
        self.EquipmentContainer = None
        self.name = ''
        self.p = 0
        self.q = 0


class Breaker(DiPole):

    def __init__(self, rfid, tpe):
        DiPole.__init__(self, rfid, tpe)

        self.normalOpen = True
        self.retained = True
        self.EquipmentContainer = None
        self.open = False

    def get_nodes(self):
        """
        Get the TopologyNodes of this branch
        :return: two TopologyNodes or nothing
        """
        try:
            terminals = list(self.references_to_me['Terminal'])

            if len(terminals) == 2:
                n1 = terminals[0].TopologicalNode
                n2 = terminals[1].TopologicalNode
                return n1, n2
            else:
                return None, None

        except KeyError:
            return None, None


class Switch(DiPole):

    def __init__(self, rfid, tpe):
        DiPole.__init__(self, rfid, tpe)

        self.normalOpen = True
        self.retained = True
        self.EquipmentContainer = None
        self.open = False
        self.description = ''
        self.aggregate = False
        self.name = ''

    def get_nodes(self):
        """
        Get the TopologyNodes of this branch
        :return: two TopologyNodes or nothing
        """
        try:
            terminals = list(self.references_to_me['Terminal'])

            if len(terminals) == 2:
                n1 = terminals[0].TopologicalNode
                n2 = terminals[1].TopologicalNode
                return n1, n2
            else:
                return None, None

        except KeyError:
            return None, None


class LoadBreakSwitch(DiPole):

    def __init__(self, rfid, tpe):
        DiPole.__init__(self, rfid, tpe)

        self.normalOpen = True
        self.retained = True
        self.EquipmentContainer = None
        self.open = False
        self.description = ''
        self.aggregate = False
        self.name = ''

        self.aor = ''

        self.serviceState = ''
        self.PSRType = ''
        self.underground = True
        self.Circuit = ''
        self.costPerUnit = 0
        self.manufacturer = ''
        self.model = ''
        self.private = False
        self.failureRate = 0
        self.importance = ""
        self.ratedVoltage = 11000
        self.access = ""
        self.extendedStatus = ""
        self.isolating = True
        self.mechanism = ""
        self.normalOpen = True
        self.ratedCurrent = 0
        self.breakingCapacity = 0
        self.shortTermCurrent = 0

    def get_nodes(self):
        """
        Get the TopologyNodes of this branch
        :return: two TopologyNodes or nothing
        """
        try:
            terminals = list(self.references_to_me['Terminal'])

            if len(terminals) == 2:
                n1 = terminals[0].TopologicalNode
                n2 = terminals[1].TopologicalNode
                return n1, n2
            else:
                return None, None

        except KeyError:
            return None, None


class Line(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)


class ACLineSegment(DiPole):

    def __init__(self, rfid, tpe):
        DiPole.__init__(self, rfid, tpe)

        self.BaseVoltage = None
        self.current_limit = None
        self.bch = 0
        self.gch = 0
        self.r = 0
        self.x = 0

    def get_voltage(self):

        if self.BaseVoltage is not None:
            return self.BaseVoltage.nominalVoltage
        else:
            if 'Terminal' in self.references_to_me.keys():
                tps = list(self.references_to_me['Terminal'])

                if len(tps) > 0:
                    tp = tps[0]

                    return tp.get_voltage()
                else:
                    return None
            else:
                return None

    def get_pu_values(self, Sbase=100):
        """
        Get the per-unit values of the equivalent PI model
        :param Sbase: Sbase in MVA
        :return: R, X, Gch, Bch
        """
        if self.BaseVoltage is not None:
            Vnom = self.get_voltage()

            if Vnom is not None:

                Zbase = (Vnom * Vnom) / Sbase
                Ybase = 1.0 / Zbase

                # at this point r, x, g, b are the complete values for all the line length
                R = self.r / Zbase
                X = self.x / Zbase
                G = self.gch / Ybase
                B = self.bch / Ybase
            else:
                R = 0
                X = 0
                G = 0
                B = 0
        else:
            R = 0
            X = 0
            G = 0
            B = 0

        return R, X, G, B

    def get_rate(self):
        return 1e-20

    def get_dict(self):
        d = super().get_dict()
        R, X, G, B = self.get_pu_values(Sbase=100)
        v = self.get_voltage()
        d['r'] = self.r
        d['x'] = self.x
        d['g'] = self.gch
        d['b'] = self.bch
        d['r_pu'] = R
        d['x_pu'] = X
        d['g_pu'] = G
        d['b_pu'] = B
        d['Vnom'] = v if v is not None else ''
        return d


class PowerTransformerEnd(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)

        self.BaseVoltage = None
        self.b = 0
        self.g = 0
        self.r = 0
        self.x = 0
        self.ratedS = 0
        self.ratedU = 0
        self.PowerTransformer = None
        self.endNumber = 0
        self.connectionKind = ""
        self.Terminal = None

    def get_voltage(self):

        if self.ratedU > 0:
            return self.ratedU
        else:
            if self.BaseVoltage is not None:
                return self.BaseVoltage.nominalVoltage
            else:
                return None

    def get_pu_values(self, Sbase_system=100):
        """
        Get the per-unit values of the equivalent PI model
        :return: R, X, Gch, Bch
        """
        if self.ratedS > 0 and self.ratedU > 0:
            Zbase = (self.ratedU * self.ratedU) / self.ratedS
            Ybase = 1.0 / Zbase
            machine_to_sys = Sbase_system / self.ratedS
            # at this point r, x, g, b are the complete values for all the line length
            R = self.r / Zbase * machine_to_sys
            X = self.x / Zbase * machine_to_sys
            G = self.g / Ybase * machine_to_sys
            B = self.b / Ybase * machine_to_sys
        else:
            R = 0
            X = 0
            G = 0
            B = 0

        return R, X, G, B

    def get_dict(self):
        d = super().get_dict()
        R, X, G, B = self.get_pu_values()

        d['ratedS'] = self.ratedS
        d['ratedU'] = self.ratedU
        d['endNumber'] = self.endNumber
        d['r'] = self.r
        d['x'] = self.x
        d['g'] = self.g
        d['b'] = self.b
        d['r_pu'] = R
        d['x_pu'] = X
        d['g_pu'] = G
        d['b_pu'] = B
        return d


class PowerTransformer(DiPole):

    def __init__(self, rfid, tpe):
        DiPole.__init__(self, rfid, tpe)

        self.EquipmentContainer = None

    def get_windings_number(self):
        """
        Get the number of windings
        :return: # number of associated windings
        """
        try:
            return len(self.references_to_me['PowerTransformerEnd'])
        except KeyError:
            return 0

    def get_windings(self):
        """
        Get list of windings
        :return: list of winding objects
        """
        try:
            return list(self.references_to_me['PowerTransformerEnd'])
        except KeyError:
            return list()

    def get_pu_values(self):
        """
        Get the transformer p.u. values
        :return:
        """
        try:
            windings = self.get_windings()

            if len(windings) == 2:
                R, X, G, B = 0, 0, 0, 0
                for winding in windings:
                    r, x, g, b = winding.get_pu_values()
                    R += r
                    X += x
                    G += g
                    B += b
            else:
                R, X, G, B = 0, 0, 0, 0

        except KeyError:
            R, X, G, B = 0, 0, 0, 0

        return R, X, G, B

    def get_voltages(self):
        """

        :return:
        """
        return [x.get_voltage() for x in self.get_windings()]

    def get_rate(self):

        rating = 0
        for winding in self.get_windings():
            if winding.ratedS > rating:
                rating = winding.ratedS

        return rating

    def get_dict(self):
        d = super().get_dict()
        R, X, G, B = self.get_pu_values()
        voltages = self.get_voltages()
        n_windings = self.get_windings_number()
        d['r_pu'] = R
        d['x_pu'] = X
        d['g_pu'] = G
        d['b_pu'] = B
        d['windings_number'] = n_windings
        for i in range(n_windings):
            v = voltages[i]
            d['V' + str(i+1)] = v if v is not None else ''
        return d


class Winding(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)

        self.tap_changers = list()


class EnergyConsumer(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)

        self.LoadResponse = None
        self.EquipmentContainer = None
        self.name = ''
        self.p = 0
        self.q = 0


class ConformLoad(MonoPole):

    def __init__(self, rfid, tpe):
        MonoPole.__init__(self, rfid, tpe)

        self.LoadGroup = None
        self.LoadResponse = None
        self.EquipmentContainer = None
        self.p = 0
        self.q = 0

    def get_dict(self):

        d = super().get_dict()
        d['p'] = self.p
        d['q'] = self.q

        return d

    def get_pq(self):
        return self.p, self.q


class NonConformLoad(MonoPole):

    def __init__(self, rfid, tpe):
        MonoPole.__init__(self, rfid, tpe)

        self.name = ''
        self.EquipmentContainer = None
        self.LoadResponse = None
        self.LoadGroup = None

    def get_pq(self):
        return 0, 0


class NonConformLoadGroup(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)

        self.name = ''


class LoadResponseCharacteristic(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)

        self.exponentModel = True
        self.pVoltageExponent = 0
        self.qVoltageExponent = 0


class RegulatingControl(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)

        self.mode = ''
        self.Terminal = None
        self.discrete = False
        self.enabled = True
        self.targetDeadband = 0
        self.targetValue = 0
        self.targetValueUnitMultiplier = ''


class RatioTapChanger(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)

        self.stepVoltageIncrement = 0
        self.tculControlMode = ''
        self.TransformerEnd = None
        self.RatioTapChangerTable = None
        self.highStep = 0
        self.lowStep = 0
        self.ltcFlag = False
        self.neutralStep = 0
        self.neutralU = 0
        self.normalStep = 0
        self.TapChangerControl = None
        self.name = ''
        self.controlEnabled = True
        self.step = 0


class GeneratingUnit(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)

        self.initialP = 0
        self.longPF = 0
        self.maxOperatingP = 0
        self.minOperatingP = 0
        self.nominalP = 0
        self.shortPF = 0
        self.EquipmentContainer = None
        self.description = ''
        self.normalPF = 0


class SynchronousMachine(MonoPole):

    def __init__(self, rfid, tpe):
        MonoPole.__init__(self, rfid, tpe)

        self.qPercent = 0
        self.type = ''
        self.InitialReactiveCapabilityCurve = None
        self.ratedS = 0
        self.GeneratingUnit = None
        self.RegulatingControl = None
        self.EquipmentContainer = None
        self.operatingMode = ''
        self.referencePriority = ''
        self.p = 0
        self.q = 0
        self.controlEnabled = True

    def get_dict(self):
        """

        :return:
        """
        d = super().get_dict()

        d['GeneratingUnit'] = self.GeneratingUnit.uuid if self.GeneratingUnit is not None else ''

        d['p'] = self.p
        d['q'] = self.q

        d['qPercent'] = self.qPercent
        d['ratedS'] = self.ratedS
        d['controlEnabled'] = self.controlEnabled
        d['type'] = self.type

        return d


class HydroGenerationUnit(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)

        self.name = ''
        self.description = ''
        self.maxOperatingP = 0
        self.minOperatingP = 0
        self.EquipmentContainer = None
        self.initialP = 0
        self.longPF = 0
        self.shortPF = 0
        self.nominalP = 0

    def get_dict(self):
        """

        :return:
        """
        d = super().get_dict()

        d['EquipmentContainer'] = self.EquipmentContainer.uuid if self.EquipmentContainer is not None else ''
        d['maxOperatingP'] = self.maxOperatingP
        d['minOperatingP'] = self.minOperatingP
        d['initialP'] = self.initialP
        d['nominalP'] = self.nominalP
        d['longPF'] = self.longPF
        d['shortPF'] = self.shortPF

        return d


class HydroPowerPlant(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)

        self.name = ''
        self.hydroPlantStorageType = ''

    def get_dict(self):
        """

        :return:
        """
        d = super().get_dict()

        d['hydroPlantStorageType'] = self.hydroPlantStorageType

        return d


class LinearShuntCompensator(MonoPole):

    def __init__(self, rfid, tpe):
        MonoPole.__init__(self, rfid, tpe)

        self.name = ''
        self.EquipmentContainer = None
        self.bPerSection = 0
        self.gPerSection = 0
        self.maximumSections = 0
        self.nomU = 0
        self.normalSections = 0

    def get_dict(self):
        """

        :return:
        """
        d = super().get_dict()

        d['EquipmentContainer'] = self.EquipmentContainer.uuid if self.EquipmentContainer is not None else ''

        d['bPerSection'] = self.bPerSection
        d['gPerSection'] = self.gPerSection
        d['maximumSections'] = self.maximumSections
        d['nomU'] = self.nomU
        d['normalSections'] = self.normalSections

        return d


class NuclearGeneratingUnit(MonoPole):

    def __init__(self, rfid, tpe):
        MonoPole.__init__(self, rfid, tpe)

        self.name = ''
        self.description = ''
        self.maxOperatingP = 0
        self.minOperatingP = 0
        self.EquipmentContainer = None
        self.initialP = 0
        self.longPF = 0
        self.shortPF = 0
        self.nominalP = 0
        self.variableCost = 0

    def get_dict(self):
        """

        :return:
        """
        d = super().get_dict()

        d['EquipmentContainer'] = self.EquipmentContainer.uuid if self.EquipmentContainer is not None else ''
        d['maxOperatingP'] = self.maxOperatingP
        d['minOperatingP'] = self.minOperatingP
        d['initialP'] = self.initialP
        d['nominalP'] = self.nominalP
        d['longPF'] = self.longPF
        d['shortPF'] = self.shortPF
        d['variableCost'] = self.variableCost

        return d


class RatioTapChangerTable(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)

        self.name = ''


class RatioTapChangerTablePoint(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)

        self.ratio = 0
        self.step = 0
        self.RatioTapChangerTable = None

    def get_dict(self):
        """

        :return:
        """
        d = super().get_dict()

        d['RatioTapChangerTable'] = self.RatioTapChangerTable.uuid if self.RatioTapChangerTable is not None else ''
        d['ratio'] = self.ratio
        d['step'] = self.step

        return d


class ReactiveCapabilityCurve(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)

        self.name = ''
        self.curveStyle = ''
        self.xUnit = ''
        self.y1Unit = ''
        self.y2Unit = ''


class StaticVarCompensator(MonoPole):

    def __init__(self, rfid, tpe):
        MonoPole.__init__(self, rfid, tpe)

        self.name = ''
        self.EquipmentContainer = None
        self.slope = 0
        self.inductiveRating = 0
        self.capacitiveRating = 0
        self.voltageSetPoint = 0
        self.sVCControlMode = ''
        self.RegulatingControl = None

    def get_dict(self):
        """

        :return:
        """
        d = super().get_dict()

        d['EquipmentContainer'] = self.EquipmentContainer.uuid if self.EquipmentContainer is not None else ''
        d['RegulatingControl'] = self.RegulatingControl.uuid if self.RegulatingControl is not None else ''
        d['slope'] = self.slope
        d['inductiveRating'] = self.inductiveRating
        d['capacitiveRating'] = self.capacitiveRating
        d['voltageSetPoint'] = self.voltageSetPoint
        d['sVCControlMode'] = self.sVCControlMode

        return d


class TapChangerControl(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)

        self.name = ''
        self.Terminal = None
        self.mode = ''


class ThermalGenerationUnit(MonoPole):

    def __init__(self, rfid, tpe):
        MonoPole.__init__(self, rfid, tpe)

        self.name = ''
        self.description = ''
        self.maxOperatingP = 0
        self.minOperatingP = 0
        self.EquipmentContainer = None
        self.initialP = 0
        self.longPF = 0
        self.shortPF = 0
        self.nominalP = 0

    def get_dict(self):
        """

        :return:
        """
        d = super().get_dict()

        d['EquipmentContainer'] = self.EquipmentContainer.uuid if self.EquipmentContainer is not None else ''
        d['maxOperatingP'] = self.maxOperatingP
        d['minOperatingP'] = self.minOperatingP
        d['initialP'] = self.initialP
        d['nominalP'] = self.nominalP
        d['longPF'] = self.longPF
        d['shortPF'] = self.shortPF

        return d


class WindGenerationUnit(MonoPole):

    def __init__(self, rfid, tpe):
        MonoPole.__init__(self, rfid, tpe)

        self.name = ''
        self.description = ''
        self.maxOperatingP = 0
        self.minOperatingP = 0
        self.EquipmentContainer = None
        self.initialP = 0
        self.longPF = 0
        self.shortPF = 0
        self.nominalP = 0
        self.variableCost = 0
        self.windGenUnitType = ''

    def get_dict(self):
        """

        :return:
        """
        d = super().get_dict()

        d['EquipmentContainer'] = self.EquipmentContainer.uuid if self.EquipmentContainer is not None else ''
        d['maxOperatingP'] = self.maxOperatingP
        d['minOperatingP'] = self.minOperatingP
        d['initialP'] = self.initialP
        d['nominalP'] = self.nominalP
        d['longPF'] = self.longPF
        d['shortPF'] = self.shortPF
        d['variableCost'] = self.variableCost
        d['windGenUnitType'] = self.windGenUnitType
        d['description'] = self.description

        return d


class FullModel(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)

        self.scenarioTime = ''
        self.created = ''
        self.version = ''
        self.profile = ''
        self.modelingAuthoritySet = ''
        self.DependentOn = ''

    def get_dict(self):
        """

        :return:
        """
        d = super().get_dict()

        d['scenarioTime'] = self.scenarioTime
        d['created'] = self.created
        d['version'] = self.version
        d['profile'] = self.profile
        d['modelingAuthoritySet'] = self.modelingAuthoritySet
        d['DependentOn'] = self.longDependentOnPF

        return d


class TieFlow(GeneralContainer):

    def __init__(self, rfid, tpe):
        GeneralContainer.__init__(self, rfid, tpe)

        self.ControlArea = None
        self.Terminal = None
        self.positiveFlowIn = True

    def get_dict(self):
        """

        :return:
        """
        d = super().get_dict()

        d['ControlArea'] = self.ControlArea.uuid if self.ControlArea is not None else ''
        d['Terminal'] = self.Terminal.uuid if self.Terminal is not None else ''
        d['positiveFlowIn'] = self.positiveFlowIn

        return d




# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import json
import pandas as pd
import numpy as np
import time
import multiprocessing
from sklearn.cluster import KMeans
from sklearn.cluster import SpectralClustering

from GridCal.Engine.basic_structures import Logger
from GridCal.Engine.Core.multi_circuit import MultiCircuit
from GridCal.Engine.Simulations.PowerFlow.power_flow_options import PowerFlowOptions
from GridCal.Engine.Core.time_series_pf_data import compile_time_circuit, BranchImpedanceMode
from GridCal.Engine.Simulations.PowerFlow.time_series_driver import TimeSeries
from GridCal.Engine.Simulations.driver_types import SimulationTypes
from GridCal.Engine.Simulations.Clustering.clustering import kmeans_approximate_sampling


class TimeSeriesClustering(TimeSeries):
    name = 'Time Series Clustering'
    tpe = SimulationTypes.ClusteringTimeSeries_run

    def __init__(self, grid: MultiCircuit, options: PowerFlowOptions, opf_time_series_results=None,
                 start_=0, end_=None, cluster_number=10):
        """
        TimeSeriesClustering constructor
        @param grid: MultiCircuit instance
        @param options: PowerFlowOptions instance
        """
        TimeSeries.__init__(self,  grid=grid, options=options, opf_time_series_results=opf_time_series_results,
                            start_=start_, end_=end_)

        self.cluster_number = cluster_number

        self.sampled_time_idx = list()
        self.sampled_probabilities = list()

    def get_steps(self):
        """
        Get time steps list of strings
        """
        return [l.strftime('%d-%m-%Y %H:%M') for l in pd.to_datetime(self.grid.time_profile[self.sampled_time_idx])]

    def run(self):
        """
        Run the time series simulation
        @return:
        """

        a = time.time()

        if self.end_ is None:
            self.end_ = len(self.grid.time_profile)
        time_indices = np.arange(self.start_, self.end_)

        # compile the multi-circuit
        time_circuit = compile_time_circuit(circuit=self.grid,
                                            apply_temperature=False,
                                            branch_tolerance_mode=BranchImpedanceMode.Specified,
                                            opf_results=self.opf_time_series_results)

        self.progress_text.emit('Clustering...')
        X = time_circuit.Sbus
        X = X[:, time_indices].real.T
        self.sampled_time_idx, self.sampled_probabilities = kmeans_approximate_sampling(X, n_points=self.cluster_number)

        self.results = self.run_single_thread(time_indices=self.sampled_time_idx)

        self.elapsed = time.time() - a




from GridCal.Engine.IO.cim.cim_parser import CIMImport, CIMExport
from GridCal.Engine.IO.dgs_parser import *
from GridCal.Engine.IO.dpx_parser import *
from GridCal.Engine.IO.ipa_parser import *
from GridCal.Engine.IO.json_parser import *
from GridCal.Engine.IO.matpower.matpower_parser import parse_matpower_file
from GridCal.Engine.IO.raw_parser import PSSeParser
from GridCal.Engine.IO.excel_interface import *
from GridCal.Engine.IO.file_handler import FileOpen, FileSave



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'gui.ui',
# licensing of 'gui.ui' applies.
#
# Created: Mon Jan  6 14:16:49 2020
#      by: pyside2-uic  running on PySide2 5.13.0
#
# WARNING! All changes made in this file will be lost!

from PySide2 import QtCore, QtGui, QtWidgets

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(941, 590)
        self.verticalLayout = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout.setContentsMargins(1, 1, 1, 1)
        self.verticalLayout.setObjectName("verticalLayout")
        self.frame = QtWidgets.QFrame(Dialog)
        self.frame.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame.setObjectName("frame")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.frame)
        self.verticalLayout_2.setContentsMargins(-1, -1, -1, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.treeView = QtWidgets.QTreeView(self.frame)
        self.treeView.setAlternatingRowColors(True)
        self.treeView.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.treeView.setAnimated(True)
        self.treeView.setObjectName("treeView")
        self.verticalLayout_2.addWidget(self.treeView)
        self.verticalLayout.addWidget(self.frame)
        self.frame_2 = QtWidgets.QFrame(Dialog)
        self.frame_2.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_2.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_2.setObjectName("frame_2")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.frame_2)
        self.horizontalLayout.setContentsMargins(-1, 0, -1, -1)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.accept_selected_pushButton = QtWidgets.QPushButton(self.frame_2)
        self.accept_selected_pushButton.setText("")
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/Icons/icons/accept.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.accept_selected_pushButton.setIcon(icon)
        self.accept_selected_pushButton.setObjectName("accept_selected_pushButton")
        self.horizontalLayout.addWidget(self.accept_selected_pushButton)
        self.reject_selected_pushButton = QtWidgets.QPushButton(self.frame_2)
        self.reject_selected_pushButton.setText("")
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/Icons/icons/delete.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.reject_selected_pushButton.setIcon(icon1)
        self.reject_selected_pushButton.setObjectName("reject_selected_pushButton")
        self.horizontalLayout.addWidget(self.reject_selected_pushButton)
        spacerItem = QtWidgets.QSpacerItem(632, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.doit_pushButton = QtWidgets.QPushButton(self.frame_2)
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/Icons/icons/accept2.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.doit_pushButton.setIcon(icon2)
        self.doit_pushButton.setObjectName("doit_pushButton")
        self.horizontalLayout.addWidget(self.doit_pushButton)
        self.verticalLayout.addWidget(self.frame_2)

        self.retranslateUi(Dialog)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QtWidgets.QApplication.translate("Dialog", "Dialog", None, -1))
        self.accept_selected_pushButton.setToolTip(QtWidgets.QApplication.translate("Dialog", "Accept selected", None, -1))
        self.reject_selected_pushButton.setToolTip(QtWidgets.QApplication.translate("Dialog", "Reject selected changes", None, -1))
        self.doit_pushButton.setToolTip(QtWidgets.QApplication.translate("Dialog", "Process all changes as especified", None, -1))
        self.doit_pushButton.setText(QtWidgets.QApplication.translate("Dialog", "Do it", None, -1))

from .icons_rc import *

if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    Dialog = QtWidgets.QDialog()
    ui = Ui_Dialog()
    ui.setupUi(Dialog)
    Dialog.show()
    sys.exit(app.exec_())




# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
import numpy as np
from typing import List
from GridCal.Gui.GuiFunctions import get_list_model
from GridCal.Gui.GridEditorWidget.generic_graphics import *
from GridCal.Gui.GridEditorWidget.bus_graphics import TerminalItem
from GridCal.Gui.GridEditorWidget.messages import *
from GridCal.Gui.GuiFunctions import BranchObjectModel
from GridCal.Engine.Devices.transformer import Transformer2W, TransformerType
from GridCal.Engine.Devices.branch import BranchType
from GridCal.Engine.Simulations.Topology.topology_driver import reduce_grid_brute


class TransformerEditor(QDialog):

    def __init__(self, branch: Transformer2W, Sbase=100, modify_on_accept=True, templates=None, current_template=None):
        """
        Transformer
        :param branch:
        :param Sbase:
        """
        super(TransformerEditor, self).__init__()

        # keep pointer to the line object
        self.transformer_obj = branch

        self.Sbase = Sbase

        self.modify_on_accept = modify_on_accept

        self.templates = self.filter_valid_templates(templates)

        self.current_template = current_template

        self.selected_template = None

        self.setObjectName("self")

        self.setContextMenuPolicy(Qt.NoContextMenu)

        self.layout = QVBoxLayout(self)

        # ------------------------------------------------------------------------------------------
        # Set the object values
        # ------------------------------------------------------------------------------------------
        self.Vf = self.transformer_obj.bus_from.Vnom
        self.Vt = self.transformer_obj.bus_to.Vnom

        # Change the impedances to the system base
        base_change = Sbase / (self.transformer_obj.rate + 1e-9)

        R = self.transformer_obj.R / base_change
        X = self.transformer_obj.X / base_change
        G = self.transformer_obj.G / base_change
        B = self.transformer_obj.B / base_change
        Sn = self.transformer_obj.rate

        zsc = np.sqrt(R * R + X * X)
        Vsc = 100.0 * zsc
        Pcu = R * Sn * 1000.0

        if abs(G) > 0.0 and abs(B) > 0.0:
            zl = 1.0 / complex(G, B)
            rfe = zl.real
            xm = zl.imag

            Pfe = 1000.0 * Sn / rfe

            k = 1 / (rfe * rfe) + 1 / (xm * xm)
            I0 = 100.0 * np.sqrt(k)
        else:
            Pfe = 0
            I0 = 0

        # ------------------------------------------------------------------------------------------

        # catalogue
        self.catalogue_combo = QComboBox()
        if templates is not None:
            if len(self.templates) > 0:

                self.catalogue_combo.setModel(get_list_model(self.templates))

                if self.current_template is not None:
                    try:
                        idx = self.templates.index(self.current_template)
                        self.catalogue_combo.setCurrentIndex(idx)

                        # set the template parameters
                        Sn = self.current_template.rating  # MVA
                        Pcu = self.current_template.Pcu  # kW
                        Pfe = self.current_template.Pfe  # kW
                        I0 = self.current_template.I0  # %
                        Vsc = self.current_template.Vsc  # %
                    except:
                        pass

        # load template
        self.load_template_btn = QPushButton()
        self.load_template_btn.setText('Load template values')
        self.load_template_btn.clicked.connect(self.load_template_btn_click)

        # Sn
        self.sn_spinner = QDoubleSpinBox()
        self.sn_spinner.setMinimum(0)
        self.sn_spinner.setMaximum(9999999)
        self.sn_spinner.setDecimals(6)
        self.sn_spinner.setValue(Sn)

        # Pcu
        self.pcu_spinner = QDoubleSpinBox()
        self.pcu_spinner.setMinimum(0)
        self.pcu_spinner.setMaximum(9999999)
        self.pcu_spinner.setDecimals(6)
        self.pcu_spinner.setValue(Pcu)

        # Pfe
        self.pfe_spinner = QDoubleSpinBox()
        self.pfe_spinner.setMinimum(0)
        self.pfe_spinner.setMaximum(9999999)
        self.pfe_spinner.setDecimals(6)
        self.pfe_spinner.setValue(Pfe)

        # I0
        self.I0_spinner = QDoubleSpinBox()
        self.I0_spinner.setMinimum(0)
        self.I0_spinner.setMaximum(9999999)
        self.I0_spinner.setDecimals(6)
        self.I0_spinner.setValue(I0)

        # Vsc
        self.vsc_spinner = QDoubleSpinBox()
        self.vsc_spinner.setMinimum(0)
        self.vsc_spinner.setMaximum(9999999)
        self.vsc_spinner.setDecimals(6)
        self.vsc_spinner.setValue(Vsc)

        # accept button
        self.accept_btn = QPushButton()
        self.accept_btn.setText('Accept')
        self.accept_btn.clicked.connect(self.accept_click)

        # labels

        # add all to the GUI

        # add all to the GUI
        if templates is not None:
            self.layout.addWidget(QLabel("Suitable templates"))
            self.layout.addWidget(self.catalogue_combo)
            self.layout.addWidget(self.load_template_btn)
            self.layout.addWidget(QLabel(""))

        self.layout.addWidget(QLabel("Sn: Nominal power [MVA]"))
        self.layout.addWidget(self.sn_spinner)

        self.layout.addWidget(QLabel("Pcu: Copper losses [kW]"))
        self.layout.addWidget(self.pcu_spinner)

        self.layout.addWidget(QLabel("Pfe: Iron losses [kW]"))
        self.layout.addWidget(self.pfe_spinner)

        self.layout.addWidget(QLabel("I0: No load current [%]"))
        self.layout.addWidget(self.I0_spinner)

        self.layout.addWidget(QLabel("Vsc: Short circuit voltage [%]"))
        self.layout.addWidget(self.vsc_spinner)

        # self.layout.addWidget(self.system_base_chk)

        self.layout.addWidget(self.accept_btn)

        self.setLayout(self.layout)

        self.setWindowTitle('Transformer editor')

    def filter_valid_templates(self, templates: List[TransformerType]):
        """
        Filter templates
        :param templates:
        :return:
        """
        if templates is None:
            return None

        lst = list()

        Vf = self.transformer_obj.bus_from.Vnom
        Vt = self.transformer_obj.bus_to.Vnom

        for tpe in templates:

            HV2 = tpe.HV * 1.01
            HV1 = tpe.HV * 0.99

            LV2 = tpe.LV * 1.01
            LV1 = tpe.LV * 0.99

            # check that the voltages are within a 1% tolerance
            if (HV1 < Vf < HV2) or (LV1 < Vf < LV2):
                if (HV1 < Vt < HV2) or (LV1 < Vt < LV2):
                    lst.append(tpe)

        return lst

    def get_template(self):
        """
        Fabricate template values from the branch values
        :return: TransformerType instance
        """
        eps = 1e-20
        Vf = self.transformer_obj.bus_from.Vnom  # kV
        Vt = self.transformer_obj.bus_to.Vnom  # kV
        Sn = self.sn_spinner.value() + eps  # MVA
        Pcu = self.pcu_spinner.value() + eps  # kW
        Pfe = self.pfe_spinner.value() + eps  # kW
        I0 = self.I0_spinner.value() + eps  # %
        Vsc = self.vsc_spinner.value()  # %

        Pfe = eps if Pfe == 0.0 else Pfe
        I0 = eps if I0 == 0.0 else I0

        tpe = TransformerType(hv_nominal_voltage=Vf,
                              lv_nominal_voltage=Vt,
                              nominal_power=Sn,
                              copper_losses=Pcu,
                              iron_losses=Pfe,
                              no_load_current=I0,
                              short_circuit_voltage=Vsc,
                              gr_hv1=0.5,
                              gx_hv1=0.5)

        return tpe

    def accept_click(self):
        """
        Create transformer type and get the impedances
        :return:
        """

        if self.modify_on_accept:

            if self.selected_template is None:
                # no selected template, but a new one was generated
                tpe = self.get_template()
            else:
                # pick the last selected template
                tpe = self.selected_template

            self.transformer_obj.apply_template(tpe, Sbase=self.Sbase)

        self.accept()

    def load_template(self, template: TransformerType):
        """

        :param template:
        :return:
        """
        self.sn_spinner.setValue(template.rating)  # MVA
        self.pcu_spinner.setValue(template.Pcu)  # kW
        self.pfe_spinner.setValue(template.Pfe)  # kW
        self.I0_spinner.setValue(template.I0)  # %
        self.vsc_spinner.setValue(template.Vsc)  # %

        self.selected_template = template

    def load_template_btn_click(self):
        """
        Accept template values
        """

        if self.templates is not None:

            idx = self.catalogue_combo.currentIndex()
            template = self.templates[idx]

            if isinstance(template, TransformerType):
                self.load_template(template)


class TransformerGraphicItem(QGraphicsLineItem):

    def __init__(self, fromPort: TerminalItem, toPort: TerminalItem, diagramScene, width=5,
                 branch: Transformer2W = None):
        """

        :param fromPort:
        :param toPort:
        :param diagramScene:
        :param width:
        :param branch:
        """
        QGraphicsLineItem.__init__(self, None)

        self.api_object = branch
        if self.api_object is not None:
            if self.api_object.active:
                self.style = ACTIVE['style']
                self.color = ACTIVE['color']
            else:
                self.style = DEACTIVATED['style']
                self.color = DEACTIVATED['color']
        else:
            self.style = OTHER['style']
            self.color = OTHER['color']
        self.width = width
        self.pen_width = width
        self.setPen(QPen(self.color, self.width, self.style))
        self.setFlag(self.ItemIsSelectable, True)
        self.setCursor(QCursor(Qt.PointingHandCursor))

        self.pos1 = None
        self.pos2 = None
        self.fromPort = None
        self.toPort = None
        self.diagramScene = diagramScene

        if fromPort:
            self.setFromPort(fromPort)

        if toPort:
            self.setToPort(toPort)

        # add transformer circles
        self.symbol_type = BranchType.Line
        self.symbol = None
        self.c0 = None
        self.c1 = None
        self.c2 = None
        if self.api_object is not None:
            self.update_symbol()

        # add the line and it possible children to the scene
        self.diagramScene.addItem(self)

        if fromPort and toPort:
            self.redraw()

    def set_colour(self, color: QColor, w, style: Qt.PenStyle):
        """
        Set color and style
        :param color: QColor instance
        :param w: width
        :param style: PenStyle instance
        :return:
        """
        self.setPen(QPen(color, w, style))
        self.c2.setPen(QPen(color, w, style))
        self.c1.setPen(QPen(color, w, style))

    def remove_symbol(self):
        """
        Remove all symbols
        """
        for elm in [self.symbol, self.c1, self.c2, self.c0]:
            if elm is not None:
                try:
                    self.diagramScene.removeItem(elm)
                    # sip.delete(elm)
                    elm = None
                except:
                    pass

    def update_symbol(self):
        """
        Make the branch symbol
        :return:
        """

        # remove the symbol of the branch
        self.remove_symbol()
        self.make_transformer_symbol()
        self.symbol_type = BranchType.Transformer

    def make_transformer_symbol(self):
        """
        create the transformer simbol
        :return:
        """
        h = 80.0
        w = h
        d = w/2
        self.symbol = QGraphicsRectItem(QRectF(0, 0, w, h), parent=self)
        self.symbol.setPen(QPen(Qt.transparent))

        self.c0 = QGraphicsEllipseItem(0, 0, d, d, parent=self.symbol)
        self.c1 = QGraphicsEllipseItem(0, 0, d, d, parent=self.symbol)
        self.c2 = QGraphicsEllipseItem(0, 0, d, d, parent=self.symbol)

        self.c0.setPen(QPen(Qt.transparent, self.width, self.style))
        self.c2.setPen(QPen(self.color, self.width, self.style))
        self.c1.setPen(QPen(self.color, self.width, self.style))

        self.c0.setBrush(QBrush(Qt.white))
        self.c2.setBrush(QBrush(Qt.white))

        self.c0.setPos(w * 0.35 - d / 2, h * 0.5 - d / 2)
        self.c1.setPos(w * 0.35 - d / 2, h * 0.5 - d / 2)
        self.c2.setPos(w * 0.65 - d / 2, h * 0.5 - d / 2)

        self.c0.setZValue(0)
        self.c1.setZValue(2)
        self.c2.setZValue(1)

    def setToolTipText(self, toolTip: str):
        """
        Set branch tool tip text
        Args:
            toolTip: text
        """
        self.setToolTip(toolTip)

        if self.symbol is not None:
            self.symbol.setToolTip(toolTip)

        if self.c0 is not None:
            self.c0.setToolTip(toolTip)
            self.c1.setToolTip(toolTip)
            self.c2.setToolTip(toolTip)

    def contextMenuEvent(self, event):
        """
        Show context menu
        @param event:
        @return:
        """
        if self.api_object is not None:
            menu = QMenu()
            menu.addSection("Transformer")

            pe = menu.addAction('Active')
            pe.setCheckable(True)
            pe.setChecked(self.api_object.active)
            pe.triggered.connect(self.enable_disable_toggle)

            ra2 = menu.addAction('Delete')
            del_icon = QIcon()
            del_icon.addPixmap(QPixmap(":/Icons/icons/delete3.svg"))
            ra2.setIcon(del_icon)
            ra2.triggered.connect(self.remove)

            re = menu.addAction('Reduce')
            re_icon = QIcon()
            re_icon.addPixmap(QPixmap(":/Icons/icons/grid_reduction.svg"))
            re.setIcon(re_icon)
            re.triggered.connect(self.reduce)

            ra3 = menu.addAction('Editor')
            edit_icon = QIcon()
            edit_icon.addPixmap(QPixmap(":/Icons/icons/edit.svg"))
            ra3.setIcon(edit_icon)
            ra3.triggered.connect(self.edit)

            ra6 = menu.addAction('Plot profiles')
            plot_icon = QIcon()
            plot_icon.addPixmap(QPixmap(":/Icons/icons/plot.svg"))
            ra6.setIcon(plot_icon)
            ra6.triggered.connect(self.plot_profiles)

            ra3 = menu.addAction('Add to catalogue')
            ra3_icon = QIcon()
            ra3_icon.addPixmap(QPixmap(":/Icons/icons/Catalogue.svg"))
            ra3.setIcon(ra3_icon)
            ra3.triggered.connect(self.add_to_templates)

            ra4 = menu.addAction('Assign rate to profile')
            ra4_icon = QIcon()
            ra4_icon.addPixmap(QPixmap(":/Icons/icons/assign_to_profile.svg"))
            ra4.setIcon(ra4_icon)
            ra4.triggered.connect(self.assign_rate_to_profile)

            ra5 = menu.addAction('Assign active state to profile')
            ra5_icon = QIcon()
            ra5_icon.addPixmap(QPixmap(":/Icons/icons/assign_to_profile.svg"))
            ra5.setIcon(ra5_icon)
            ra5.triggered.connect(self.assign_status_to_profile)

            ra7 = menu.addAction('Flip')
            ra7_icon = QIcon()
            ra7_icon.addPixmap(QPixmap(":/Icons/icons/redo.svg"))
            ra7.setIcon(ra7_icon)
            ra7.triggered.connect(self.flip_connections)

            menu.addSection('Tap changer')

            ra4 = menu.addAction('Tap up')
            ra4_icon = QIcon()
            ra4_icon.addPixmap(QPixmap(":/Icons/icons/up.svg"))
            ra4.setIcon(ra4_icon)
            ra4.triggered.connect(self.tap_up)

            ra5 = menu.addAction('Tap down')
            ra5_icon = QIcon()
            ra5_icon.addPixmap(QPixmap(":/Icons/icons/down.svg"))
            ra5.setIcon(ra5_icon)
            ra5.triggered.connect(self.tap_down)

            menu.exec_(event.screenPos())

        else:
            pass

    def mousePressEvent(self, QGraphicsSceneMouseEvent):
        """
        mouse press: display the editor
        :param QGraphicsSceneMouseEvent:
        :return:
        """

        mdl = BranchObjectModel([self.api_object], self.api_object.editable_headers,
                                parent=self.diagramScene.parent().object_editor_table,
                                editable=True, transposed=True,
                                non_editable_attributes=self.api_object.non_editable_attributes)

        self.diagramScene.parent().object_editor_table.setModel(mdl)

    def mouseDoubleClickEvent(self, event):
        """
        On double click, edit
        :param event:
        :return:
        """

        if self.api_object.branch_type in [BranchType.Transformer, BranchType.Line]:
            # trigger the editor
            self.edit()
        elif self.api_object.branch_type is BranchType.Switch:
            # change state
            self.enable_disable_toggle()

    def remove(self, ask=True):
        """
        Remove this object in the diagram and the API
        @return:
        """
        if ask:
            ok = yes_no_question('Do you want to remove this transformer?', 'Remove transformer')
        else:
            ok = True

        if ok:
            self.diagramScene.circuit.delete_branch(self.api_object)
            self.diagramScene.removeItem(self)

    def reduce(self):
        """
        Reduce this branch
        """
        ok = yes_no_question('Do you want to reduce this transformer?', 'Reduce transformer')

        if ok:
            # get the index of the branch
            br_idx = self.diagramScene.circuit.branches.index(self.api_object)

            # call the reduction routine
            removed_branch, removed_bus, \
                updated_bus, updated_branches = reduce_grid_brute(self.diagramScene.circuit, br_idx)

            # remove the reduced branch
            removed_branch.graphic_obj.remove_symbol()
            self.diagramScene.removeItem(removed_branch.graphic_obj)

            # update the buses (the deleted one and the updated one)
            if removed_bus is not None:
                # merge the removed bus with the remaining one
                updated_bus.graphic_obj.merge(removed_bus.graphic_obj)

                # remove the updated bus children
                for g in updated_bus.graphic_obj.shunt_children:
                    self.diagramScene.removeItem(g.nexus)
                    self.diagramScene.removeItem(g)
                # re-draw the children
                updated_bus.graphic_obj.create_children_icons()

                # remove bus
                for g in removed_bus.graphic_obj.shunt_children:
                    self.diagramScene.removeItem(g.nexus)  # remove the links between the bus and the children
                self.diagramScene.removeItem(removed_bus.graphic_obj)  # remove the bus and all the children contained

                #
                # updated_bus.graphic_obj.update()

            for br in updated_branches:
                # remove the branch from the schematic
                self.diagramScene.removeItem(br.graphic_obj)
                # add the branch to the schematic with the rerouting and all
                self.diagramScene.parent_.add_transformer(br)
                # update both buses
                br.bus_from.graphic_obj.update()
                br.bus_to.graphic_obj.update()

    def remove_widget(self):
        """
        Remove this object in the diagram
        @return:
        """
        self.diagramScene.removeItem(self)

    def enable_disable_toggle(self):
        """

        @return:
        """
        if self.api_object is not None:
            if self.api_object.active:
                self.set_enable(False)
            else:
                self.set_enable(True)

            if self.diagramScene.circuit.has_time_series:
                ok = yes_no_question('Do you want to update the time series active status accordingly?',
                                     'Update time series active status')

                if ok:
                    # change the bus state (time series)
                    self.diagramScene.set_active_status_to_profile(self.api_object, override_question=True)

    def set_enable(self, val=True):
        """
        Set the enable value, graphically and in the API
        @param val:
        @return:
        """
        self.api_object.active = val
        if self.api_object is not None:
            if self.api_object.active:
                self.style = ACTIVE['style']
                self.color = ACTIVE['color']
            else:
                self.style = DEACTIVATED['style']
                self.color = DEACTIVATED['color']
        else:
            self.style = OTHER['style']
            self.color = OTHER['color']

        # Switch coloring
        if self.symbol_type == BranchType.Switch:
            if self.api_object.active:
                self.symbol.setBrush(self.color)
            else:
                self.symbol.setBrush(Qt.white)

        if self.symbol_type == BranchType.DCLine:
            self.symbol.setBrush(self.color)
            if self.api_object.active:
                self.symbol.setPen(QPen(ACTIVE['color']))
            else:
                self.symbol.setPen(QPen(DEACTIVATED['color']))

        # Set pen for everyone
        self.set_pen(QPen(self.color, self.width, self.style))

    def plot_profiles(self):
        """
        Plot the time series profiles
        @return:
        """
        i = self.diagramScene.circuit.get_branches().index(self.api_object)
        self.diagramScene.plot_branch(i, self.api_object)

    def setFromPort(self, fromPort):
        """
        Set the From terminal in a connection
        @param fromPort:
        @return:
        """
        self.fromPort = fromPort
        if self.fromPort:
            self.pos1 = fromPort.scenePos()
            self.fromPort.posCallbacks.append(self.setBeginPos)
            self.fromPort.parent.setZValue(0)

    def setToPort(self, toPort):
        """
        Set the To terminal in a connection
        @param toPort:
        @return:
        """
        self.toPort = toPort
        if self.toPort:
            self.pos2 = toPort.scenePos()
            self.toPort.posCallbacks.append(self.setEndPos)
            self.toPort.parent.setZValue(0)

    def setEndPos(self, endpos):
        """
        Set the starting position
        @param endpos:
        @return:
        """
        self.pos2 = endpos
        self.redraw()

    def setBeginPos(self, pos1):
        """
        Set the starting position
        @param pos1:
        @return:
        """
        self.pos1 = pos1
        self.redraw()

    def redraw(self):
        """
        Redraw the line with the given positions
        @return:
        """
        if self.pos1 is not None and self.pos2 is not None:

            # Set position
            self.setLine(QLineF(self.pos1, self.pos2))

            # set Z-Order (to the back)
            self.setZValue(-1)

            if self.api_object is not None:

                # if the object branch type is different from the current displayed type, change it
                if self.symbol_type != self.api_object.branch_type:
                    self.update_symbol()

                if self.api_object.branch_type == BranchType.Line:
                    pass

                elif self.api_object.branch_type == BranchType.Branch:
                    pass

                else:

                    # if the branch has a moveable symbol, move it
                    try:
                        h = self.pos2.y() - self.pos1.y()
                        b = self.pos2.x() - self.pos1.x()
                        ang = np.arctan2(h, b)
                        h2 = self.symbol.rect().height() / 2.0
                        w2 = self.symbol.rect().width() / 2.0
                        a = h2 * np.cos(ang) - w2 * np.sin(ang)
                        b = w2 * np.sin(ang) + h2 * np.cos(ang)

                        center = (self.pos1 + self.pos2) * 0.5 - QPointF(a, b)

                        transform = QTransform()
                        transform.translate(center.x(), center.y())
                        transform.rotate(np.rad2deg(ang))
                        self.symbol.setTransform(transform)

                    except Exception as ex:
                        print(ex)

    def set_pen(self, pen):
        """
        Set pen to all objects
        Args:
            pen:
        """
        self.setPen(pen)

        # Color the symbol only for switches
        if self.api_object.branch_type == BranchType.Switch:
            if self.symbol is not None:
                self.symbol.setPen(pen)

        elif self.api_object.branch_type == BranchType.Transformer:
            if self.c1 is not None:
                self.c1.setPen(pen)
                self.c2.setPen(pen)

    def edit(self):
        """
        Open the appropriate editor dialogue
        :return:
        """
        Sbase = self.diagramScene.circuit.Sbase
        templates = self.diagramScene.circuit.transformer_types
        current_template = self.api_object.template
        dlg = TransformerEditor(self.api_object, Sbase,
                                modify_on_accept=True,
                                templates=templates,
                                current_template=current_template)
        if dlg.exec_():
            pass

    def add_to_templates(self):
        """
        Open the appropriate editor dialogue
        :return:
        """
        Sbase = self.diagramScene.circuit.Sbase

        if self.api_object.template is not None:
            # automatically pick the template
            if isinstance(self.api_object.template, TransformerType):
                self.diagramScene.circuit.add_transformer_type(self.api_object.template)
            else:
                # raise dialogue to set the template
                dlg = TransformerEditor(self.api_object, Sbase, modify_on_accept=False)
                if dlg.exec_():
                    tpe = dlg.get_template()
                    self.diagramScene.circuit.add_transformer_type(tpe)
        else:
            # raise dialogue to set the template
            dlg = TransformerEditor(self.api_object, Sbase, modify_on_accept=False)
            if dlg.exec_():
                tpe = dlg.get_template()
                self.diagramScene.circuit.add_transformer_type(tpe)

    def assign_rate_to_profile(self):
        """
        Assign the snapshot rate to the profile
        """
        self.diagramScene.set_rate_to_profile(self.api_object)

    def flip_connections(self):
        """
        Flip connections
        :return:
        """
        self.api_object.flip()

    def assign_status_to_profile(self):
        """
        Assign the snapshot rate to the profile
        """
        self.diagramScene.set_active_status_to_profile(self.api_object)

    def tap_up(self):
        """
        Set one tap up
        """
        self.api_object.tap_up()

    def tap_down(self):
        """
        Set one tap down
        """
        self.api_object.tap_down()

